#!/usr/bin/env python
# flake8: noqa: C0301

"""Tests for creating signed transactions."""

"""We're going to have to do this the hard way and make up some utxos."""


import unittest
from zpywallet.address.btc import BitcoinAddress
from zpywallet.address.eth import EthereumAddress
from zpywallet.destination import Destination
from zpywallet.network import BitcoinMainNet, BitcoinSegwitMainNet, EthereumMainNet
from zpywallet.transactions.encode import create_transaction
from zpywallet.utils.keys import PrivateKey, PublicKey
from zpywallet.utxo import UTXO
from zpywallet.nodes.btc import btc_nodes
from zpywallet.nodes.eth import eth_nodes
from zpywallet.transactions.decode import transaction_size_simple

class TestAddress(unittest.TestCase):
    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_legacy_sign(self):
        """Test creating Satoshi-like legacy transactions."""
        # To make things clear, we will use fake UTXOs from this address,
        # derived from private key 0, which nobody can spend.
        # We will use a fake private key (1) since we do not need to broadcast
        # it anywhere, and that particular functionality has its own unit test.
        b = BitcoinAddress(['16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM'])
        b.sync()
        saved_utxos = b.get_utxos()
        destinations = [Destination("16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM", 0.00000001, BitcoinMainNet),
                        Destination("1BgGZ9tcN4rm9KBzDn7KprQz87SZ26SAMH", 0.00000002, BitcoinMainNet)]
        # The (1) private key has a sweeper attached to it so its balanace should always be zero.
        # Therefore, the wallet.create_transaction method should fail with not enough funds.
        utxos = []
        for u in saved_utxos:
            _u = UTXO(None, None, _unsafe_internal_testing_only={'amount': u.amount})
            _u._output['amount'] = u.amount
            _u._output['address'] = u.address
            _u._output['height'] = u.height
            _u._output['confirmed'] = u.confirmed
            _u._output['txid'] = u.txid
            _u._output['index'] = u.index
            _u._output['private_key'] = PrivateKey.from_int(1).to_wif()
            _u._output['script_pubkey'] = PublicKey.script(u.address, BitcoinMainNet)
            utxos.append(_u)
        if len(utxos) > 0:
            create_transaction(utxos, destinations, network=BitcoinMainNet, full_nodes=btc_nodes)

    def test_001_fake_segwit_sign(self):
        """Test creating Satoshi-like segwit transactions which have no segwit inputs, so f=alling back to legacy signing."""
        # To make things clear, we will use fake UTXOs from this address,
        # derived from private key 0, which nobody can spend.
        # We will use a fake private key (1) since we do not need to broadcast
        # it anywhere, and that particular functionality has its own unit test.
        b = BitcoinAddress(['16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM'])
        b.sync()
        saved_utxos = b.get_utxos()
        destinations = [Destination("16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM", 0.00000001, BitcoinSegwitMainNet),
                        Destination("1BgGZ9tcN4rm9KBzDn7KprQz87SZ26SAMH", 0.00000002, BitcoinSegwitMainNet)]
        # The (1) private key has a sweeper attached to it so its balanace should always be zero.
        # Therefore, the wallet.create_transaction method should fail with not enough funds.
        utxos = []
        for u in saved_utxos:
            _u = UTXO(None, None, _unsafe_internal_testing_only={'amount': u.amount})
            _u._output['amount'] = u.amount
            _u._output['address'] = u.address
            _u._output['height'] = u.height
            _u._output['confirmed'] = u.confirmed
            _u._output['txid'] = u.txid
            _u._output['index'] = u.index
            _u._output['private_key'] = PrivateKey.from_int(1).to_wif()
            _u._output['script_pubkey'] = PublicKey.script(u.address, BitcoinSegwitMainNet)
            utxos.append(_u)
        if len(utxos) > 0:
            create_transaction(utxos, destinations, network=BitcoinSegwitMainNet, full_nodes=btc_nodes)

    def test_002_segwit_sign(self):
        """Test creating Satoshi-like segwit transactions, all segwit inputs."""
        # To make things clear, we will use fake UTXOs from this address,
        # derived from private key 0, which nobody can spend.
        # We will use a fake private key (1) since we do not need to broadcast
        # it anywhere, and that particular functionality has its own unit test.
        bc = BitcoinAddress(['bc1q34aq5drpuwy3wgl9lhup9892qp6svr8ldzyy7c'])
        bc.sync()
        saved_utxos = bc.get_utxos()
        destinations = [Destination("16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM", 0.00000001, BitcoinSegwitMainNet),
                        Destination("1BgGZ9tcN4rm9KBzDn7KprQz87SZ26SAMH", 0.00000002, BitcoinSegwitMainNet)]
        # The (1) private key has a sweeper attached to it so its balanace should always be zero.
        # Therefore, the wallet.create_transaction method should fail with not enough funds
        utxos = []
        for u in saved_utxos:
            _u = UTXO(None, None, _unsafe_internal_testing_only={'amount': u.amount})
            _u._output['amount'] = u.amount
            _u._output['address'] = u.address
            _u._output['height'] = u.height
            _u._output['confirmed'] = u.confirmed
            _u._output['txid'] = u.txid
            _u._output['index'] = u.index
            _u._output['private_key'] = PrivateKey.from_int(1).to_wif()
            _u._output['script_pubkey'] = PublicKey.script(u.address, BitcoinSegwitMainNet)
            utxos.append(_u)
        if len(utxos) > 0:
            create_transaction(utxos, destinations, network=BitcoinSegwitMainNet, full_nodes=btc_nodes)
    
    def test_003_segwit_sign_partial(self):
        """Test creating Satoshi-like segwit transactions, mixed segwit and legacy inputs."""
        # To make things clear, we will use fake UTXOs from this address,
        # derived from private key 0, which nobody can spend.
        # We will use a fake private key (1) since we do not need to broadcast
        # it anywhere, and that particular functionality has its own unit test.
        bc2 = BitcoinAddress(['16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM', 'bc1q34aq5drpuwy3wgl9lhup9892qp6svr8ldzyy7c'])
        bc2.sync()
        saved_utxos = bc2.get_utxos()
        destinations = [Destination("16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM", 0.00000001, BitcoinSegwitMainNet),
                        Destination("1BgGZ9tcN4rm9KBzDn7KprQz87SZ26SAMH", 0.00000002, BitcoinSegwitMainNet)]
        # The (1) private key has a sweeper attached to it so its balanace should always be zero.
        # Therefore, the wallet.create_transaction method should fail with not enough funds.
        utxos = []
        for u in saved_utxos:
            _u = UTXO(None, None, _unsafe_internal_testing_only={'amount': u.amount})
            _u._output['amount'] = u.amount
            _u._output['address'] = u.address
            _u._output['height'] = u.height
            _u._output['confirmed'] = u.confirmed
            _u._output['txid'] = u.txid
            _u._output['index'] = u.index
            _u._output['private_key'] = PrivateKey.from_int(1).to_wif()
            _u._output['script_pubkey'] = PublicKey.script(u.address, BitcoinSegwitMainNet)
            utxos.append(_u)
        if len(utxos) > 0:
            create_transaction(utxos, destinations, network=BitcoinSegwitMainNet, full_nodes=btc_nodes)

    def test_004_sign_with_change(self):
        """Test creating Satoshi-like transactions with change calculation"""
        # To make things clear, we will use fake UTXOs from this address,
        # derived from private key 0, which nobody can spend.
        # We will use a fake private key (1) since we do not need to broadcast
        # it anywhere, and that particular functionality has its own unit test.
        bc2 = BitcoinAddress(['16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM', 'bc1q34aq5drpuwy3wgl9lhup9892qp6svr8ldzyy7c'])
        bc2.sync()
        saved_utxos = bc2.get_utxos()
        destinations = [Destination("16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM", 0.00000001, BitcoinSegwitMainNet),
                        Destination("1BgGZ9tcN4rm9KBzDn7KprQz87SZ26SAMH", 0.00000002, BitcoinSegwitMainNet)]
        # The (1) private key has a sweeper attached to it so its balanace should always be zero.
        # Therefore, the wallet.create_transaction method should fail with not enough funds.
        utxos = []
        for u in saved_utxos:
            _u = UTXO(None, None, _unsafe_internal_testing_only={'amount': u.amount})
            _u._output['amount'] = u.amount
            _u._output['address'] = u.address
            _u._output['height'] = u.height
            _u._output['confirmed'] = u.confirmed
            _u._output['txid'] = u.txid
            _u._output['index'] = u.index
            _u._output['private_key'] = PrivateKey.from_int(1).to_wif()
            _u._output['script_pubkey'] = PublicKey.script(u.address, BitcoinSegwitMainNet)
            utxos.append(_u)
        if len(utxos) > 0:
            temp_transaction = create_transaction(utxos, destinations, network=BitcoinSegwitMainNet, full_nodes=btc_nodes)
            fee_rate = 1
            size = transaction_size_simple(temp_transaction)
            total_inputs = sum([i.amount(in_standard_units=False) for i in utxos])
            total_outputs = sum([o.amount(in_standard_units=False) for o in destinations])
            if total_inputs < total_outputs + size*fee_rate:
                raise ValueError("Not enough balance for this transaction")
            change_value = total_inputs - total_outputs - size*fee_rate
            if change_value > 0:
                change = Destination('16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM', change_value / 1e8, BitcoinSegwitMainNet)
                destinations.append(change)
                create_transaction(utxos, destinations, network=BitcoinSegwitMainNet, full_nodes=btc_nodes)

    def test_005_eth_sign(self):
        """Test creating EVM Ethereum transactions"""
        b = EthereumAddress(['0xd73e8e2ac0099169e7404f23c6caa94cf1884384'])
        b.sync()
        destinations = [Destination("0xea83c649dd49a6ec44c9e2943eb673a8fbb7bab6", 0.00000002, EthereumMainNet)]

        utxos = []
        _u = UTXO(None, None, _unsafe_internal_testing_only={'amount': b.get_balance()[0]})
        _u._output['address'] = "0xd73e8e2ac0099169e7404f23c6caa94cf1884384"
        _u._output['private_key'] = "0x0000000000000000000000000000000000000000000000000000000000000001"
        utxos.append(_u)
        # There seems to be a bug where the stock ETH nodes do not allow connections from inside the Github Actions runner or at least from Tox
        try:
            create_transaction(utxos, destinations, network=EthereumMainNet, full_nodes=eth_nodes, gas=1, gasPrice=1)
        except RuntimeError as e:
            pass



# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
btc_nodes = [
    {
        'url': 'https://rpc.coinsdo.net/btc',
    },
    {
        'url': 'https://bitcoin-mainnet-archive.allthatnode.com/'
    },
    {
        'url': 'https://bitcoin.drpc.org/'
    },
    {
        'url': 'https://bitcoin-api.flare.network'
    },
]

btc_esplora_nodes = [
    {
        'url': 'https://mempool.ninja'
    },
    {
        'url': 'https://www.lohkoketju.fi/'
    },
]


import time
import requests

from ...errors import NetworkException
from ...generated import wallet_pb2

class EsploraAddress:
    """
    A class representing a Bitcoin address.

    This class allows you to retrieve the balance and transaction history of a Bitcoin address using the Esplora API.
    Esplora is deployed on many popular websites, including mempool.space (Rate limited) and blockstream.info.

    Note: Esplora has a built-in limitation of returning up to 50 unconfirmed transactions per address. While this should be
    large enough for most use cases, if you run into problems, try using a different address provider.

    Note 2: This API will not return the Genesis block in transactions, unlike the other balances. This will affect
    balance displayed for Satoshi's first address 1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa - but that output is unspendable anyway.

    Args:
        address (str): The human-readable Bitcoin address.

    Attributes:
        address (str): The human-readable Bitcoin address.
        endpoint (str): The Esplora endpoint to use. Defaults to Blockstream's endpoint.
        transaction_history (list): The cached list of transactions.
        height (int): The last known block height.

    Methods:
        get_balance(): Retrieves the total and confirmed balances of the Bitcoin address.
        get_utxos(): Retrieves the UTXO set for this address
        get_block_height(): Retrieves the current block height.
        get_transaction_history(): Retrieves the transaction history of the Bitcoin address.

    Raises:
        Exception: If the API request fails or the address balance/transaction history cannot be retrieved.
    """

    def _clean_tx(self, element):
        new_element = wallet_pb2.Transaction()
        new_element.txid = element['txid']
        if 'block_height' in element['status'].keys():
            new_element.confirmed = True
            new_element.height = element['status']['block_height']
        else:
            new_element.confirmed = False
            
        if new_element.confirmed:
            new_element.timestamp = element['status']['block_time']

        for vin in element['vin']:
            txinput = new_element.btclike_transaction.inputs.add()
            txinput.txid = vin['txid']
            txinput.index = vin['vout']
            if vin['prevout'] is not None:
                txinput.amount = int(vin['prevout']['value'])
            else:
                txinput.amount = int(0)

        i = 0
        for vout in element['vout']:
            txoutput = new_element.btclike_transaction.outputs.add()
            txoutput.amount = int(vout['value'])
            txoutput.index = i
            i += 1
            if 'scriptpubkey_address' in vout.keys():
                txoutput.address = vout['scriptpubkey_address']

        # Now we must calculate the total fee
        total_inputs = sum([a.amount for a in new_element.btclike_transaction.inputs])
        total_outputs = sum([a.amount for a in new_element.btclike_transaction.outputs])

        new_element.total_fee = (total_inputs - total_outputs)

        new_element.btclike_transaction.fee = int(element['fee'])
        new_element.fee_metric = wallet_pb2.VBYTE
        return new_element

    def __init__(self, addresses, request_interval=(3,1), transactions=None, **kwargs):
        """
        Initializes an instance of the EsploraAddress class.

        Args:
            addresses (list): A list of human-readable Bitcoin addresses.
            endpoint (str): The Esplora endpoint to use.
            request_interval (tuple): A pair of integers indicating the number of requests allowed during
                a particular amount of seconds. Set to (0,N) for no rate limiting, where N>0.
        """
        # exampBlockstream.info's rate limits are unknown.
        # Ostensibly there are no limits for that site, but I got 429 errors when testing with (1000,1), so
        # the default limit will be the same as for mempool.space - 3 requests per second.
        self.requests, self.interval_sec = request_interval
        self.addresses = addresses
        self.endpoint = kwargs.get('url')
        if transactions is not None and isinstance(transactions, list):
            self.transactions = transactions
        else:
            self.transactions = []

    def sync(self):
        self.transactions = [*self._get_transaction_history()]
        self.height = self.get_block_height()

    def get_balance(self):
        """
        Retrieves the balance of the Bitcoin address.

        Returns:
            float: The balance of the Bitcoin address in BTC.

        Raises:
            Exception: If the API request fails or the address balance cannot be retrieved.
        """
        utxos = self.get_utxos()
        total_balance = 0
        confirmed_balance = 0
        for utxo in utxos:
            total_balance += utxo.amount
            if utxo.confirmed:
                confirmed_balance += utxo.amount
        return total_balance, confirmed_balance
        
    def get_utxos(self):
        # Transactions are generated in reverse order
        utxos = []
        for i in range(len(self.transactions)-1, -1, -1):
            for out in self.transactions[i].outputs:
                if out.spent:
                    continue
                if out.address in self.addresses:
                    utxo = wallet_pb2.UTXO()
                    utxo.address = out.address
                    utxo.txid = self.transactions[i].txid
                    utxo.index = out.index
                    utxo.amount = out.amount
                    utxo.height = self.transactions[i].height
                    utxo.confirmed = self.transactions[i].confirmed
                    utxos.append(utxo)
        return utxos
    
    def get_block_height(self):
        # Get the current block height now:
        url = f"{self.endpoint}/blocks/tip/height"
        for attempt in range(3, -1, -1):
            if attempt == 0:
                raise NetworkException("Network request failure")
            try:
                response = requests.get(url, timeout=60)
                break
            except requests.RequestException:
                pass
            except requests.exceptions.JSONDecodeError:
                pass
        
        if response.status_code == 200:
            self.height = int(response.text)
            return self.height
        else:
            try:
                return self.height
            except AttributeError as exc:
                raise NetworkException("Failed to retrieve current blockchain height") from exc

        
    def get_transaction_history(self):
        """
        Retrieves the transaction history of the Bitcoin address from cached data augmented with network data.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        if len(self.transactions) == 0:
            self.transactions = [*self._get_transaction_history()]
        else:
            # First element is the most recent transactions
            txhash = self.transactions[0].txid
            txs = [*self._get_transaction_history(txhash)]
            txs.extend(self.transactions)
            self.transactions = txs
                    
        return self.transactions

    def _get_transaction_history(self, txhash=None):
        """
        Retrieves the transaction history of the Bitcoin address.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        for address in self.addresses:
            # This gets up to 50 mempool transactions + up to 25 confirmed transactions
            url = f"{self.endpoint}/address/{address}/txs"
            for attempt in range(3, -1, -1):
                if attempt == 0:
                    raise NetworkException("Network request failure")
                try:
                    response = requests.get(url, timeout=60)
                    if response.status_code == 200:
                        data = response.json()
                        break
                    else:
                        raise NetworkException("Failed to retrieve transaction history")
                except requests.RequestException:
                    pass

            if response.status_code == 200:
                data = response.json()
                for tx in data:
                    time.sleep(self.interval_sec/(self.requests*len(data)))
                    if txhash and tx["txid"] == txhash:
                        return
                    yield self._clean_tx(tx)
            else:
                raise NetworkException("Failed to retrieve transaction history")
            
            if len(data) > 0:
                last_tx = data[-1]["txid"]
            else:
                break
            
            while len(data) > 0:
                url = f"{self.endpoint}/address/{address}/txs/chain/{last_tx}"
                for attempt in range(3, -1, -1):
                    if attempt == 0:
                        raise NetworkException("Network request failure")
                    try:
                        response = requests.get(url, timeout=60)
                        break
                    except requests.RequestException:
                        pass

                if response.status_code == 200:
                    data = response.json()
                    for tx in data:
                        time.sleep(self.interval_sec/(self.requests*len(data)))
                        if txhash and tx["hash"] == txhash:
                            return
                        yield self._clean_tx(tx)
                    if len(data) > 0:
                        last_tx = data[-1]["txid"]
                else:
                    raise NetworkException("Failed to retrieve transaction history")




import requests
from ...errors import NetworkException

async def broadcast_transaction_eth_mew(raw_transaction_hex):
    api_url = "https://api.mewapi.io/v1/transaction/sendRaw"

    payload = {
        "rawTx": raw_transaction_hex,
    }

    try:
        response = requests.post(api_url, json=payload, timeout=30)
    except Exception as e:
        raise NetworkException("Connection error while broadcasting transaction: {}".format(str(e)))
    result = response.json()

    if response.status_code >= 300 and result.get("status") == "1":
        raise NetworkException(f"Failed to broadcast Ethereum transaction using MyEtherWallet: {result.get('message')}")



import hashlib
import binascii
import scrypt

from .utils.keys import PrivateKey
from .utils.utils import encrypt, decrypt
from .utils import base58

class Bip38PrivateKey:
    BLOCK_SIZE = 16
    KEY_LEN = 32
    IV_LEN = 16

    def __init__(self, privkey: PrivateKey, passphrase, compressed=True, segwit=False, witness_version=0):
        '''BIP0038 non-ec-multiply encryption. Returns BIP0038 encrypted privkey.'''
        if "BASE58" not in privkey.network.ADDRESS_MODE and "BECH32" not in privkey.network.ADDRESS_MODE:
            raise ValueError("BIP38 requires Base58 or Bech32 addresses")
        flagbyte = b'\xe0' if compressed else b'\xc0'
        addr = privkey.public_key.bech32_address(compressed, witness_version) if segwit else privkey.public_key.base58_address(compressed)
        addresshash = hashlib.sha256(hashlib.sha256(addr.encode()).digest()).digest()[0:4]
        key = scrypt.hash(passphrase, addresshash, 16384, 8, 8)
        derivedhalf1 = key[0:32]
        derivedhalf2 = key[32:64]
        encryptedhalf1 = encrypt(binascii.unhexlify('%0.32x' % (int(binascii.hexlify(bytes(privkey)[0:16]), 16) ^ int(binascii.hexlify(derivedhalf1[0:16]), 16))), derivedhalf2)
        encryptedhalf2 = encrypt(binascii.unhexlify('%0.32x' % (int(binascii.hexlify(bytes(privkey)[16:32]), 16) ^ int(binascii.hexlify(derivedhalf1[16:32]), 16))), derivedhalf2)
        self.flagbyte = flagbyte
        self.addresshash = addresshash
        self.encryptedhalf1 = encryptedhalf1
        self.encryptedhalf2 = encryptedhalf2
        encrypted_privkey = b'\x01\x42' + self.flagbyte + self.addresshash + self.encryptedhalf1 + self.encryptedhalf2
        encrypted_privkey += hashlib.sha256(hashlib.sha256(encrypted_privkey).digest()).digest()[:4] # b58check for encrypted privkey
        self._encrypted_privkey = base58.b58encode(encrypted_privkey)

    @property
    def base58(self):
        return self._encrypted_privkey.decode()
        

    def private_key(self, passphrase, compressed=True, segwit=False, witness_version=0):
        '''BIP0038 non-ec-multiply decryption. Returns WIF privkey.'''
        d = base58.b58decode(self._encrypted_privkey)
        d = d[2:]
        #flagbyte = d[0:1]
        d = d[1:]
        #WIF compression
        #if flagbyte == b'\xc0':
        #    compressed = False
        #if flagbyte == b'\xe0':
        #    compressed = True
        addresshash = d[0:4]
        d = d[4:-4]
        key = scrypt.hash(passphrase,addresshash, 16384, 8, 8)
        derivedhalf1 = key[0:32]
        derivedhalf2 = key[32:64]
        encryptedhalf1 = d[0:16]
        encryptedhalf2 = d[16:32]
        decryptedhalf2 = decrypt(encryptedhalf2, derivedhalf2)
        decryptedhalf1 = decrypt(encryptedhalf1, derivedhalf2)
        priv = decryptedhalf1 + decryptedhalf2
        priv = PrivateKey.from_bytes(binascii.unhexlify('%064x' % (int(binascii.hexlify(priv), 16) ^ int(binascii.hexlify(derivedhalf1), 16))))
        pub = priv.public_key

        addr = pub.bech32_address(compressed, witness_version) if segwit else pub.base58_address(compressed)
        if hashlib.sha256(hashlib.sha256(addr.encode()).digest()).digest()[0:4] != addresshash:
            raise ValueError('Verification failed. Password is incorrect.')
        else:
            return priv



import asyncio
import binascii
import hashlib
from .blockcypher import *

def tx_hash_bcy(raw_transaction_hex):
    return binascii.hexlify(hashlib.sha256(hashlib.sha256(raw_transaction_hex.decode()).digest()).digest())

async def broadcast_transaction_bcy(raw_transaction_hex):

    tasks = []

    tasks.append(asyncio.create_task(broadcast_transaction_bcy_blockcypher(raw_transaction_hex)))
    
    try:
        await asyncio.gather(*tasks, return_exceptions=True)
    except Exception as e:
        pass


#!/usr/bin/env python3
#
# Copyright (c) 2013 Pavol Rusnak
# Copyright (c) 2017 mruddy
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import json
import random
import unittest
from typing import List

from zpywallet.mnemonic import Mnemonic


class MnemonicTest(unittest.TestCase):
    def _check_list(self, language: str, vectors: List[str]) -> None:
        mnemo = Mnemonic(language)
        for v in vectors:
            code = mnemo.to_mnemonic(bytes.fromhex(v[0]))
            seed = Mnemonic.to_seed(code, passphrase="TREZOR")
            xprv = Mnemonic.to_hd_master_key(seed)
            self.assertIs(mnemo.check(v[1]), True)
            self.assertEqual(v[1], code)
            self.assertEqual(v[2], seed.hex())
            self.assertEqual(v[3], xprv)

    def test_vectors(self) -> None:
        with open("vectors.json", "r") as f:
            vectors = json.load(f)
        for lang in vectors.keys():
            self._check_list(lang, vectors[lang])

    def test_failed_checksum(self) -> None:
        code = (
            "bless cloud wheel regular tiny venue bird web grief security dignity zoo"
        )
        mnemo = Mnemonic("english")
        self.assertFalse(mnemo.check(code))

    def test_detection(self) -> None:
        self.assertEqual("english", Mnemonic.detect_language("security"))

        with self.assertRaises(Exception):
            Mnemonic.detect_language(
                "jaguar xxxxxxx"
            )  # Unrecognized in any known language

        with self.assertRaises(Exception):
            Mnemonic.detect_language(
                "jaguar jaguar"
            )  # Ambiguous after examining all words

        self.assertEqual("english", Mnemonic.detect_language("jaguar security"))
        self.assertEqual("french", Mnemonic.detect_language("jaguar aboyer"))

    def test_utf8_nfkd(self) -> None:
        # The same sentence in various UTF-8 forms
        words_nfkd = "Pr\u030ci\u0301s\u030cerne\u030c z\u030clut\u030couc\u030cky\u0301 ku\u030an\u030c u\u0301pe\u030cl d\u030ca\u0301belske\u0301 o\u0301dy za\u0301ker\u030cny\u0301 uc\u030cen\u030c be\u030cz\u030ci\u0301 pode\u0301l zo\u0301ny u\u0301lu\u030a"
        words_nfc = "P\u0159\xed\u0161ern\u011b \u017elu\u0165ou\u010dk\xfd k\u016f\u0148 \xfap\u011bl \u010f\xe1belsk\xe9 \xf3dy z\xe1ke\u0159n\xfd u\u010de\u0148 b\u011b\u017e\xed pod\xe9l z\xf3ny \xfal\u016f"
        words_nfkc = "P\u0159\xed\u0161ern\u011b \u017elu\u0165ou\u010dk\xfd k\u016f\u0148 \xfap\u011bl \u010f\xe1belsk\xe9 \xf3dy z\xe1ke\u0159n\xfd u\u010de\u0148 b\u011b\u017e\xed pod\xe9l z\xf3ny \xfal\u016f"
        words_nfd = "Pr\u030ci\u0301s\u030cerne\u030c z\u030clut\u030couc\u030cky\u0301 ku\u030an\u030c u\u0301pe\u030cl d\u030ca\u0301belske\u0301 o\u0301dy za\u0301ker\u030cny\u0301 uc\u030cen\u030c be\u030cz\u030ci\u0301 pode\u0301l zo\u0301ny u\u0301lu\u030a"

        passphrase_nfkd = (
            "Neuve\u030cr\u030citelne\u030c bezpec\u030cne\u0301 hesli\u0301c\u030cko"
        )
        passphrase_nfc = "Neuv\u011b\u0159iteln\u011b bezpe\u010dn\xe9 hesl\xed\u010dko"
        passphrase_nfkc = (
            "Neuv\u011b\u0159iteln\u011b bezpe\u010dn\xe9 hesl\xed\u010dko"
        )
        passphrase_nfd = (
            "Neuve\u030cr\u030citelne\u030c bezpec\u030cne\u0301 hesli\u0301c\u030cko"
        )

        seed_nfkd = Mnemonic.to_seed(words_nfkd, passphrase_nfkd)
        seed_nfc = Mnemonic.to_seed(words_nfc, passphrase_nfc)
        seed_nfkc = Mnemonic.to_seed(words_nfkc, passphrase_nfkc)
        seed_nfd = Mnemonic.to_seed(words_nfd, passphrase_nfd)

        self.assertEqual(seed_nfkd, seed_nfc)
        self.assertEqual(seed_nfkd, seed_nfkc)
        self.assertEqual(seed_nfkd, seed_nfd)

    def test_to_entropy(self) -> None:
        data = [bytes(random.getrandbits(8) for _ in range(32)) for _ in range(1024)]
        data.append(b"Lorem ipsum dolor sit amet amet.")
        m = Mnemonic("english")
        for d in data:
            self.assertEqual(m.to_entropy(m.to_mnemonic(d).split()), d)

    def test_expand_word(self) -> None:
        m = Mnemonic("english")
        self.assertEqual("", m.expand_word(""))
        self.assertEqual(" ", m.expand_word(" "))
        self.assertEqual("access", m.expand_word("access"))  # word in list
        self.assertEqual(
            "access", m.expand_word("acce")
        )  # unique prefix expanded to word in list
        self.assertEqual("acb", m.expand_word("acb"))  # not found at all
        self.assertEqual("acc", m.expand_word("acc"))  # multi-prefix match
        self.assertEqual("act", m.expand_word("act"))  # exact three letter match
        self.assertEqual(
            "action", m.expand_word("acti")
        )  # unique prefix expanded to word in list

    def test_expand(self) -> None:
        m = Mnemonic("english")
        self.assertEqual("access", m.expand("access"))
        self.assertEqual(
            "access access acb acc act action", m.expand("access acce acb acc act acti")
        )



import asyncio
import binascii
import hashlib
from .bitaps import *
from .blockchain_info import *
from .blockchair import *
from .blockcypher import *
from .blockstream import *
from .esplora import *
from .fullnode import *
from .mempool_space import *
from .smartbit import *
from .viabtc import *
from ...nodes.btc import *

def tx_hash_btc(raw_transaction_hex):
    return binascii.hexlify(hashlib.sha256(hashlib.sha256(raw_transaction_hex.decode()).digest()).digest())

async def broadcast_transaction_btc(raw_transaction_hex, **kwargs):
    rpc_nodes = kwargs.get('rpc_nodes') or []
    esplora_nodes = kwargs.get('esplora_nodes') or []

    tasks = []

    tasks.append(asyncio.create_task(broadcast_transaction_btc_bitaps(raw_transaction_hex)))
    tasks.append(asyncio.create_task(broadcast_transaction_btc_blockchain_info(raw_transaction_hex)))
    tasks.append(asyncio.create_task(broadcast_transaction_btc_blockchair(raw_transaction_hex)))
    tasks.append(asyncio.create_task(broadcast_transaction_btc_blockcypher(raw_transaction_hex)))
    tasks.append(asyncio.create_task(broadcast_transaction_btc_blockstream(raw_transaction_hex)))
    tasks.append(asyncio.create_task(broadcast_transaction_btc_mempool_space(raw_transaction_hex)))
    tasks.append(asyncio.create_task(broadcast_transaction_btc_smartbit(raw_transaction_hex)))
    tasks.append(asyncio.create_task(broadcast_transaction_btc_viabtc(raw_transaction_hex)))
    for node in rpc_nodes:
        tasks.append(asyncio.create_task(broadcast_transaction_btc_full_node(raw_transaction_hex, **node)))
    for node in btc_nodes:
        tasks.append(asyncio.create_task(broadcast_transaction_btc_full_node(raw_transaction_hex, **node)))
    for node in esplora_nodes:
        tasks.append(asyncio.create_task(broadcast_transaction_btc_esplora(raw_transaction_hex, **node)))
    for node in btc_esplora_nodes:
        tasks.append(asyncio.create_task(broadcast_transaction_btc_esplora(raw_transaction_hex, **node)))
    
    try:
        await asyncio.gather(*tasks, return_exceptions=True)
    except Exception as e:
        pass
    


# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
ltc_nodes = [
    {
        'url': 'https://rpc.coinsdo.net/ltc'
    },
    {
        'url': 'https://litecoin-api.flare.network/'
    },
]


import asyncio
from .bcy.all import broadcast_transaction_bcy, tx_hash_bcy
from .btc.all import broadcast_transaction_btc, tx_hash_btc
from .btctest.all import broadcast_transaction_btctest, tx_hash_btctest
from .dash.all import broadcast_transaction_dash, tx_hash_dash
from .dashtest.all import broadcast_transaction_dashtest, tx_hash_dashtest
from .doge.all import broadcast_transaction_doge, tx_hash_doge
from .dogetest.all import broadcast_transaction_dogetest, tx_hash_dogetest
from .eth.all import broadcast_transaction_eth, tx_hash_eth
from .ltc.all import broadcast_transaction_ltc, tx_hash_ltc
from .ltctest.all import broadcast_transaction_ltctest, tx_hash_ltctest
from ..network import *

def broadcast_transaction(transaction, network, **kwargs):
    if network.COIN == "BTC":
        if not network.TESTNET:
            asyncio.run(broadcast_transaction_btc(transaction, **kwargs))
        else:
            asyncio.run(broadcast_transaction_btctest(transaction, **kwargs))
    elif network.COIN == "LTC":
        if not network.TESTNET:
            asyncio.run(broadcast_transaction_ltc(transaction, **kwargs))
        else:
            broadcast_transaction_ltctest(transaction, **kwargs)
    elif network.COIN == "DASH":
        if not network.TESTNET:
            asyncio.run(broadcast_transaction_dash(transaction, **kwargs))
        else:
            broadcast_transaction_dashtest(transaction, **kwargs)
    elif network.COIN == "DOGE":
        if not network.TESTNET:
            asyncio.run(broadcast_transaction_doge(transaction, **kwargs))
        else:
            broadcast_transaction_dogetest(transaction, **kwargs)
    elif network.COIN == "ETH":
        asyncio.run(broadcast_transaction_eth(transaction, **kwargs))
    elif network.COIN == "BCY":
        asyncio.run(broadcast_transaction_bcy(transaction, **kwargs))
    else:
        raise ValueError("Cannot broadcast transaction: Unsupported network")
    

def tx_hash(transaction: bytes, network):
    if network.COIN == "BTC":
        if not network.TESTNET:
            asyncio.run(tx_hash_btc(transaction))
        else:
            asyncio.run(tx_hash_btctest(transaction))
    elif network.COIN == "LTC":
        if not network.TESTNET:
            asyncio.run(tx_hash_ltc(transaction))
        else:
            tx_hash_ltctest(transaction)
    elif network.COIN == "DASH":
        if not network.TESTNET:
            asyncio.run(tx_hash_dash(transaction))
        else:
            tx_hash_dashtest(transaction)
    elif network.COIN == "DOGE":
        if not network.TESTNET:
            asyncio.run(tx_hash_doge(transaction))
        else:
            tx_hash_dogetest(transaction)
    elif network.COIN == "ETH":
        asyncio.run(tx_hash_eth(transaction))
    elif network.COIN == "BCY":
        asyncio.run(tx_hash_bcy(transaction))
    else:
        raise ValueError("Cannot broadcast transaction: Unsupported network")


import random
from datetime import datetime
import requests

from ...errors import NetworkException
from ...utils.descriptors import descsum_create_only
from ...transactions.decode import parse_transaction_simple
from ...generated import wallet_pb2

def to_descriptor(address):
    ad = f"addr({address})"
    return f"{ad}#{descsum_create_only(ad)}"

def to_extpub_descriptor(expub, path):
    """
    Creates a descriptor out of the extended public key `expub`,
    and a path `path`.

    Parameters:
        expub (str): an extended public key. Can have any valid prefix, e.g. xpub, ypub, zpub, etc.
        path (str): a path string in the form of (for example) "m/0/0'/1". A non-hardened number
            can be replaced with '*' to fill in a range for a later RPC call. This can only be specified once.
             The path must start with 'm/'.
    
    Returns:
        str: A descriptor string.
    """
    if path[:2] != "m/":
        raise NetworkException("Path must start with 'm/")
    desc = expub + path[1:]
    return f"{desc}#{descsum_create_only(desc)}"

class BitcoinRPCClient:
    """Address querying class for Bitcoin full nodes utilizing descriptors.
       Requires node version v0.17.0 or later running with -txindex.
    """
    
    # Not static because we need to make calls to fetch input transactions.
    def _clean_tx(self, element):
        new_element = wallet_pb2.Transaction()
        new_element.txid = element['txid']
        if 'blockheight' in element.keys():
            new_element.confirmed = True
            new_element.height = element['blockheight']
        else:
            new_element.confirmed = False

        if 'blocktime' not in element.keys():
            new_element.timestamp = element['blocktime']

        element = element['decoded']

        for vin in element['vin']:
            txinput = new_element.btclike_transaction.inputs.add()
            if 'txid' in vin.keys():
                # If there is a vin txid then it's not e.g. a coinbase transaction input
                txinput.txid = vin['txid']
                txinput.index = vin['vout']
                # To fill in the amount, we have to get the other transaction id
                # But only if we're not parsing a coinbase transaction
                res = self._send_rpc_request('getrawtransaction', params=[vin['txid']])
                rawtx = res['result']
                fine_rawtx, _ = parse_transaction_simple(rawtx)
                txinput.amount = int(fine_rawtx["outputs"][txinput.index]["value"])

        for vout in element['vout']:
            txoutput = new_element.btclike_transaction.outputs.add()
            txoutput.amount = int(vout['value'])
            txoutput.index = vout['n']
            if 'address' in vout['scriptPubKey']:
                txoutput.address = vout['scriptPubKey']['address']
        
        # Now we must calculate the total fee
        total_inputs = sum([a.amount for a in new_element.btclike_transaction.inputs])
        total_outputs = sum([a.amount for a in new_element.btclike_transaction.outputs])

        new_element.total_fee = (total_inputs - total_outputs)

        new_element.btclike_transaction.fee = int((total_inputs - total_outputs) // element['vsize'])
        new_element.fee_metric = wallet_pb2.VBYTE
        return new_element

    def __init__(self, addresses, last_update=0, transactions=None, **kwargs):
        self.rpc_url = kwargs.get('url')
        self.rpc_user = kwargs.get('user')
        self.rpc_password = kwargs.get('password')
        self.client_number = kwargs.get('client_number') or 0
        self.user_id = kwargs.get('user_id') or 0
        self.max_tx_at_once = kwargs.get('max_tx_at_once') or 1000
        
        self.transactions = []
        self.addresses = addresses
        self.last_update = last_update
        if transactions is not None and isinstance(transactions, list):
            self.transactions = transactions
        else:
            self.transactions = []
    
    def sync(self):
        self.height = self.get_block_height()
        self._load_addresses()
        self.transactions = [*self._get_transaction_history()]
    
    def _send_rpc_request(self, method, params=None, as_wallet=False):
        payload = {
            'method': method,
            'params': params or [],
            'jsonrpc': '2.0',
            'id': random.randint(1, 999999)
        }
        # Full nodes wallet RPC requests are notoriously slow if there are many transactions in the node.
        try:
            response = requests.post(f"{self.rpc_url}/wallet/zpywallet_{self.client_number}_{self.user_id}" if as_wallet \
                                     else self.rpc_url, auth=(self.rpc_user, self.rpc_password) if self.rpc_user and \
                                        self.rpc_password else None, json=payload, timeout=86400)
            j = response.json()
            if 'result' not in j.keys():
                raise NetworkException("Failed to get result")
        except Exception as e:
            raise NetworkException(f"RPC call failed: {str(e)}")
    
    def get_block_height(self):
        response = self._send_rpc_request('getblockchaininfo')
        try:
            return response['result']['blocks']
        except Exception as e:
            raise NetworkException(f"Failed to make RPC Call: {str(e)}")

    def _load_addresses(self):
        # Create a new temporary wallet
        self._send_rpc_request('createwallet', params=[f"zpywallet_{self.client_number}_{self.user_id}", True])
        
        block_height = self.height
        block_height = min(block_height, self.get_block_height())


        params = [*map(lambda at: {"desc": to_descriptor(at), "timestamp": self.last_update}, self.addresses)]

        try:
            # Load the temporary wallet
            self._send_rpc_request('loadwallet', params=[f"zpywallet_{self.client_number}_{self.user_id}"])

            self.last_update = datetime.now().timestamp()
            
            # Import the address into the wallet
            self._send_rpc_request('importdescriptors', params=[params], as_wallet=True)
        
        finally:
            # Unload the temporary wallet
            self._send_rpc_request('unloadwallet', params=[f"zpywallet_{self.client_number}_{self.user_id}"])
    
    def get_balance(self):
        """
        Retrieves the balance of the Bitcoin address.

        Returns:
            float: The balance of the Bitcoin address in BTC.

        Raises:
            Exception: If the API request fails or the address balance cannot be retrieved.
        """
        utxos = self.get_utxos()
        total_balance = 0
        confirmed_balance = 0
        for utxo in utxos:
            total_balance += utxo["amount"]
            # Careful: Block height #0 is the Genesis block - don't want to exclude that.
            # (Not that it returns it ever though!)
            if utxo["confirmed"]:
                confirmed_balance += utxo["amount"]
        return total_balance, confirmed_balance
        
    def get_utxos(self):
        self.height = self.get_block_height()
        # Transactions are generated in reverse order
        utxos = []
        for i in range(len(self.transactions)-1, -1, -1):
            for utxo in [u for u in utxos]:
                # Check if any utxo has been spent in this transaction
                for vin in self.transactions[i].inputs:
                    if vin.spent or (vin.txid == utxo["txid"] and vin["index"] == utxo.index):
                        # Spent
                        utxos.remove(utxo)
            for out in self.transactions[i].outputs:
                if out.address in self.addresses:
                    utxo = wallet_pb2.UTXO()
                    utxo.address = out.address
                    utxo.txid = self.transactions[i].txid
                    utxo.index = out.index
                    utxo.amount = out.amount
                    utxo.height = self.transactions[i].height
                    utxo.confirmed = self.transactions[i].confirmed
                    utxos.append(utxo)
        return utxos
    
    def get_transaction_history(self):
        """
        Retrieves the transaction history of the Bitcoin address from cached data augmented with network data.
        Does not include Genesis blocks.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the RPC request fails or the transaction history cannot be retrieved.
        """
        if len(self.transactions) == 0:
            self.transactions = [*self._get_transaction_history()]
        else:
            # First element is the most recent transactions
            txhash = self.transactions[0].txid
            txs = [*self._get_transaction_history(txhash)]
            txs.extend(self.transactions)
            self.transactions = txs
                
    def _get_transaction_history(self, txhash=None):
        try:
            # Load the temporary wallet
            self._send_rpc_request('loadwallet', params=[f"zpywallet_{self.client_number}_{self.user_id}"])
            
             # Get the number of transactions
            total_transactions = self._send_rpc_request('getwalletinfo', as_wallet=True)['result']['txcount']

            skip_later_txs = 0
            while skip_later_txs < total_transactions:

                # listtransactions by default return only 10 most recent transactions)
                result = self._send_rpc_request('listtransactions', params=["*", self.max_tx_at_once, skip_later_txs, True], as_wallet=True)
                transactions = result['result']
                
                # Extract the transaction IDs so we can query them verbosely
                # because the tx doesn't include info on other addresses
                for tx in transactions:
                    txid = tx['txid']
                    if txid == txhash:
                        return
                    result = self._send_rpc_request('gettransaction', params=[txid, True, True], as_wallet=True)
                    result_tx = result['result']
                    yield self._clean_tx(result_tx)
            

                # Get the number of transactions in case new, incoming txs arrived.
                # We will just skip those txs and they can be received in the next method invocation.
                total_transactions_new = self._send_rpc_request('getwalletinfo', as_wallet=True)['result']['txcount']
                if total_transactions_new > total_transactions:
                    skip_later_txs += total_transactions_new - total_transactions
                    total_transactions = total_transactions_new

                # And finally prepare to process next most recent batch of transactions
                skip_later_txs += self.max_tx_at_once
        
        finally:
            # Unload and delete the temporary wallet
            self._send_rpc_request('unloadwallet', params=[f"zpywallet_{self.client_number}_{self.user_id}"])




from ._version import __version__

from .wallet import generate_mnemonic, create_wallet, create_keypair, Wallet

from .transaction import Transaction
from .utxo import UTXO
from .destination import Destination

from .utils.bip32 import HDWallet
from .utils.keys import (
    PrivateKey, PublicKey, Point
)


__all__ = [
    'errors',
    'address',
    'mnemonic',
    'network',
    'wallet',
    'utils',
    'generated',
    'broadcast',
    'transactions',
    'nodes',
    'bip38',
    'utxo',
    'transaction',
    'destination'
]



#!/usr/bin/env python
# flake8: noqa: C0301

"""Tests for `zpywallet` package."""

import binascii
import unittest
from zpywallet import wallet
from zpywallet.bip38 import Bip38PrivateKey
from zpywallet.network import BitcoinSegwitMainNet
from zpywallet.utils.bip32 import HDWallet
from zpywallet.errors import IncompatibleNetworkException
from zpywallet.utils.keys import PrivateKey


class TestZPyWallet(unittest.TestCase):
    """Tests for `zpywallet` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_create_wallet(self):
        """Test wallet creation."""
        mne = wallet.generate_mnemonic(strength=128)
        assert mne.count(' ') == 11

        hdw = wallet.create_wallet(mnemonic="nature story debris circle decrease post "
                                    "gesture cute burger chef silent grief",
                                    network=BitcoinSegwitMainNet)
        assert hdw.serialize_b58(private=True) == \
            "xprv9s21ZrQH143K26AwKW3o1D9uVbbYJp4RCLWCVQNfarQ59cgG8udwc5SVGxrTPNo2HXLNK8ZJoavsNYuExeMmdmmGVZZh5M77UV6qbMisVSm"

    def test_001_bip32(self):
        """Test BIP32 confirmance."""

        hdw = HDWallet.from_master_seed(binascii.unhexlify("000102030405060708090a0b0c0d0e0f"), network=BitcoinSegwitMainNet)
        assert hdw.serialize_b58(private=False) == "xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2gZ29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8"
        assert hdw.serialize_b58(private=True) == "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"

        child1 = hdw.get_child(0, is_prime=True)
        assert child1.serialize_b58(private=False) == "xpub68Gmy5EdvgibQVfPdqkBBCHxA5htiqg55crXYuXoQRKfDBFA1WEjWgP6LHhwBZeNK1VTsfTFUHCdrfp1bgwQ9xv5ski8PX9rL2dZXvgGDnw"
        assert child1.serialize_b58(private=True) == "xprv9uHRZZhk6KAJC1avXpDAp4MDc3sQKNxDiPvvkX8Br5ngLNv1TxvUxt4cV1rGL5hj6KCesnDYUhd7oWgT11eZG7XnxHrnYeSvkzY7d2bhkJ7"

        child2 = child1.get_child(1, is_prime=False)
        assert child2.serialize_b58(private=False) == "xpub6ASuArnXKPbfEwhqN6e3mwBcDTgzisQN1wXN9BJcM47sSikHjJf3UFHKkNAWbWMiGj7Wf5uMash7SyYq527Hqck2AxYysAA7xmALppuCkwQ"
        assert child2.serialize_b58(private=True) == "xprv9wTYmMFdV23N2TdNG573QoEsfRrWKQgWeibmLntzniatZvR9BmLnvSxqu53Kw1UmYPxLgboyZQaXwTCg8MSY3H2EU4pWcQDnRnrVA1xe8fs"

        child3 = child2.get_child(2, is_prime=True)
        assert child3.serialize_b58(private=False) == "xpub6D4BDPcP2GT577Vvch3R8wDkScZWzQzMMUm3PWbmWvVJrZwQY4VUNgqFJPMM3No2dFDFGTsxxpG5uJh7n7epu4trkrX7x7DogT5Uv6fcLW5"
        assert child3.serialize_b58(private=True) == "xprv9z4pot5VBttmtdRTWfWQmoH1taj2axGVzFqSb8C9xaxKymcFzXBDptWmT7FwuEzG3ryjH4ktypQSAewRiNMjANTtpgP4mLTj34bhnZX7UiM"

        child4 = child3.get_child(2, is_prime=False)
        assert child4.serialize_b58(private=False) == "xpub6FHa3pjLCk84BayeJxFW2SP4XRrFd1JYnxeLeU8EqN3vDfZmbqBqaGJAyiLjTAwm6ZLRQUMv1ZACTj37sR62cfN7fe5JnJ7dh8zL4fiyLHV"
        assert child4.serialize_b58(private=True) == "xprvA2JDeKCSNNZky6uBCviVfJSKyQ1mDYahRjijr5idH2WwLsEd4Hsb2Tyh8RfQMuPh7f7RtyzTtdrbdqqsunu5Mm3wDvUAKRHSC34sJ7in334"

        child5 = child4.get_child(1000000000, is_prime=False)
        assert child5.serialize_b58(private=False) == "xpub6H1LXWLaKsWFhvm6RVpEL9P4KfRZSW7abD2ttkWP3SSQvnyA8FSVqNTEcYFgJS2UaFcxupHiYkro49S8yGasTvXEYBVPamhGW6cFJodrTHy"
        assert child5.serialize_b58(private=True) == "xprvA41z7zogVVwxVSgdKUHDy1SKmdb533PjDz7J6N6mV6uS3ze1ai8FHa8kmHScGpWmj4WggLyQjgPie1rFSruoUihUZREPSL39UNdE3BBDu76"

    def test_002_bip32(self):
        """Test BIP32 confirmance."""
        hdw = HDWallet.from_master_seed(binascii.unhexlify("fffcf9f6f3f0edeae7e4e1dedbd8d5d2cfccc9c6c3c0bdbab7b4b1aeaba8a5a29f9c999693908d8a8784817e7b7875726f6c696663605d5a5754514e4b484542"),
                                        network=BitcoinSegwitMainNet)
        assert hdw.serialize_b58(private=False) == "xpub661MyMwAqRbcFW31YEwpkMuc5THy2PSt5bDMsktWQcFF8syAmRUapSCGu8ED9W6oDMSgv6Zz8idoc4a6mr8BDzTJY47LJhkJ8UB7WEGuduB"
        assert hdw.serialize_b58(private=True) == "xprv9s21ZrQH143K31xYSDQpPDxsXRTUcvj2iNHm5NUtrGiGG5e2DtALGdso3pGz6ssrdK4PFmM8NSpSBHNqPqm55Qn3LqFtT2emdEXVYsCzC2U"

        child1 = hdw.get_child(0, is_prime=False)
        assert child1.serialize_b58(private=False) == "xpub69H7F5d8KSRgmmdJg2KhpAK8SR3DjMwAdkxj3ZuxV27CprR9LgpeyGmXUbC6wb7ERfvrnKZjXoUmmDznezpbZb7ap6r1D3tgFxHmwMkQTPH"
        assert child1.serialize_b58(private=True) == "xprv9vHkqa6EV4sPZHYqZznhT2NPtPCjKuDKGY38FBWLvgaDx45zo9WQRUT3dKYnjwih2yJD9mkrocEZXo1ex8G81dwSM1fwqWpWkeS3v86pgKt"

        child2 = child1.get_child(2147483647, is_prime=True)
        assert child2.serialize_b58(private=False) == "xpub6ASAVgeehLbnwdqV6UKMHVzgqAG8Gr6riv3Fxxpj8ksbH9ebxaEyBLZ85ySDhKiLDBrQSARLq1uNRts8RuJiHjaDMBU4Zn9h8LZNnBC5y4a"
        assert child2.serialize_b58(private=True) == "xprv9wSp6B7kry3Vj9m1zSnLvN3xH8RdsPP1Mh7fAaR7aRLcQMKTR2vidYEeEg2mUCTAwCd6vnxVrcjfy2kRgVsFawNzmjuHc2YmYRmagcEPdU9"

        child3 = child2.get_child(1, is_prime=False)
        assert child3.serialize_b58(private=False) == "xpub6DF8uhdarytz3FWdA8TvFSvvAh8dP3283MY7p2V4SeE2wyWmG5mg5EwVvmdMVCQcoNJxGoWaU9DCWh89LojfZ537wTfunKau47EL2dhHKon"
        assert child3.serialize_b58(private=True) == "xprv9zFnWC6h2cLgpmSA46vutJzBcfJ8yaJGg8cX1e5StJh45BBciYTRXSd25UEPVuesF9yog62tGAQtHjXajPPdbRCHuWS6T8XA2ECKADdw4Ef"

        child4 = child3.get_child(2147483646, is_prime=True)
        assert child4.serialize_b58(private=False) == "xpub6ERApfZwUNrhLCkDtcHTcxd75RbzS1ed54G1LkBUHQVHQKqhMkhgbmJbZRkrgZw4koxb5JaHWkY4ALHY2grBGRjaDMzQLcgJvLJuZZvRcEL"
        assert child4.serialize_b58(private=True) == "xprvA1RpRA33e1JQ7ifknakTFpgNXPmW2YvmhqLQYMmrj4xJXXWYpDPS3xz7iAxn8L39njGVyuoseXzU6rcxFLJ8HFsTjSyQbLYnMpCqE2VbFWc"

        child5 = child4.get_child(2, is_prime=False)
        assert child5.serialize_b58(private=False) == "xpub6FnCn6nSzZAw5Tw7cgR9bi15UV96gLZhjDstkXXxvCLsUXBGXPdSnLFbdpq8p9HmGsApME5hQTZ3emM2rnY5agb9rXpVGyy3bdW6EEgAtqt"
        assert child5.serialize_b58(private=True) == "xprvA2nrNbFZABcdryreWet9Ea4LvTJcGsqrMzxHx98MMrotbir7yrKCEXw7nadnHM8Dq38EGfSh6dqA9QWTyefMLEcBYJUuekgW4BYPJcr9E7j"

    def test_003_bip32(self):
        """Test BIP32 confirmance - retention of leading zeroes."""
        hdw = HDWallet.from_master_seed(binascii.unhexlify("4b381541583be4423346c643850da4b320e46a87ae3d2a4e6da11eba819cd4acba45d239319ac14f863b8d5ab5a0d0c64d2e8a1e7d1457df2e5a3c51c73235be"),
                                        network=BitcoinSegwitMainNet)
        assert hdw.serialize_b58(private=False) == "xpub661MyMwAqRbcEZVB4dScxMAdx6d4nFc9nvyvH3v4gJL378CSRZiYmhRoP7mBy6gSPSCYk6SzXPTf3ND1cZAceL7SfJ1Z3GC8vBgp2epUt13"
        assert hdw.serialize_b58(private=True) == "xprv9s21ZrQH143K25QhxbucbDDuQ4naNntJRi4KUfWT7xo4EKsHt2QJDu7KXp1A3u7Bi1j8ph3EGsZ9Xvz9dGuVrtHHs7pXeTzjuxBrCmmhgC6"

        child1 = hdw.get_child(0, is_prime=True)
        assert child1.serialize_b58(private=False) == "xpub68NZiKmJWnxxS6aaHmn81bvJeTESw724CRDs6HbuccFQN9Ku14VQrADWgqbhhTHBaohPX4CjNLf9fq9MYo6oDaPPLPxSb7gwQN3ih19Zm4Y"
        assert child1.serialize_b58(private=True) == "xprv9uPDJpEQgRQfDcW7BkF7eTya6RPxXeJCqCJGHuCJ4GiRVLzkTXBAJMu2qaMWPrS7AANYqdq6vcBcBUdJCVVFceUvJFjaPdGZ2y9WACViL4L"

    def test_004_bip32(self):
        """Test BIP32 confirmance - retention of leading zeroes."""
        hdw = HDWallet.from_master_seed(binascii.unhexlify("3ddd5602285899a946114506157c7997e5444528f3003f6134712147db19b678"),
                                        network=BitcoinSegwitMainNet)
        assert hdw.serialize_b58(private=False) == "xpub661MyMwAqRbcGczjuMoRm6dXaLDEhW1u34gKenbeYqAix21mdUKJyuyu5F1rzYGVxyL6tmgBUAEPrEz92mBXjByMRiJdba9wpnN37RLLAXa"
        assert hdw.serialize_b58(private=True) == "xprv9s21ZrQH143K48vGoLGRPxgo2JNkJ3J3fqkirQC2zVdk5Dgd5w14S7fRDyHH4dWNHUgkvsvNDCkvAwcSHNAQwhwgNMgZhLtQC63zxwhQmRv"

        child1 = hdw.get_child(0, is_prime=True)
        assert child1.serialize_b58(private=False) == "xpub69AUMk3qDBi3uW1sXgjCmVjJ2G6WQoYSnNHyzkmdCHEhSZ4tBok37xfFEqHd2AddP56Tqp4o56AePAgCjYdvpW2PU2jbUPFKsav5ut6Ch1m"
        assert child1.serialize_b58(private=True) == "xprv9vB7xEWwNp9kh1wQRfCCQMnZUEG21LpbR9NPCNN1dwhiZkjjeGRnaALmPXCX7SgjFTiCTT6bXes17boXtjq3xLpcDjzEuGLQBM5ohqkao9G"

        child2 = child1.get_child(1, is_prime=True)
        assert child2.serialize_b58(private=False) == "xpub6BJA1jSqiukeaesWfxe6sNK9CCGaujFFSJLomWHprUL9DePQ4JDkM5d88n49sMGJxrhpjazuXYWdMf17C9T5XnxkopaeS7jGk1GyyVziaMt"
        assert child2.serialize_b58(private=True) == "xprv9xJocDuwtYCMNAo3Zw76WENQeAS6WGXQ55RCy7tDJ8oALr4FWkuVoHJeHVAcAqiZLE7Je3vZJHxspZdFHfnBEjHqU5hG1Jaj32dVoS6XLT1"

    def test_005_bip32(self):
        """Test BIP32 confirmance - invalid HD keys."""
        with self.assertRaises(IncompatibleNetworkException):
            HDWallet.deserialize("xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6LBpB85b3D2yc8sfvZU521AAwdZafEz7mnzBBsz4wKY5fTtTQBm")

        with self.assertRaises(IncompatibleNetworkException):
            HDWallet.deserialize("xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFGTQQD3dC4H2D5GBj7vWvSQaaBv5cxi9gafk7NF3pnBju6dwKvH")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6Txnt3siSujt9RCVYsx4qHZGc62TG4McvMGcAUjeuwZdduYEvFn")

        with self.assertRaises(IncompatibleNetworkException):
            HDWallet.deserialize("xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFGpWnsj83BHtEy5Zt8CcDr1UiRXuWCmTQLxEK9vbz5gPstX92JQ")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6N8ZMMXctdiCjxTNq964yKkwrkBJJwpzZS4HS2fxvyYUA4q2Xe4")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFAzHGBP2UuGCqWLTAPLcMtD9y5gkZ6Eq3Rjuahrv17fEQ3Qen6J")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xprv9s2SPatNQ9Vc6GTbVMFPFo7jsaZySyzk7L8n2uqKXJen3KUmvQNTuLh3fhZMBoG3G4ZW1N2kZuHEPY53qmbZzCHshoQnNf4GvELZfqTUrcv")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xpub661no6RGEX3uJkY4bNnPcw4URcQTrSibUZ4NqJEw5eBkv7ovTwgiT91XX27VbEXGENhYRCf7hyEbWrR3FewATdCEebj6znwMfQkhRYHRLpJ")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xprv9s21ZrQH4r4TsiLvyLXqM9P7k1K3EYhA1kkD6xuquB5i39AU8KF42acDyL3qsDbU9NmZn6MsGSUYZEsuoePmjzsB3eFKSUEh3Gu1N3cqVUN")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xpub661MyMwAuDcm6CRQ5N4qiHKrJ39Xe1R1NyfouMKTTWcguwVcfrZJaNvhpebzGerh7gucBvzEQWRugZDuDXjNDRmXzSZe4c7mnTK97pTvGS8")

        with self.assertRaises(IncompatibleNetworkException):
            HDWallet.deserialize("DMwo58pR1QLEFihHiXPVykYB6fJmsTeHvyTp7hRThAtCX8CvYzgPcn8XnmdfHGMQzT7ayAmfo4z3gY5KfbrZWZ6St24UVf2Qgo6oujFktLHdHY4")

        with self.assertRaises(IncompatibleNetworkException):
            HDWallet.deserialize("DMwo58pR1QLEFihHiXPVykYB6fJmsTeHvyTp7hRThAtCX8CvYzgPcn8XnmdfHPmHJiEDXkTiJTVV9rHEBUem2mwVbbNfvT2MTcAqj3nesx8uBf9")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzF93Y5wvzdUayhgkkFoicQZcP3y52uPPxFnfoLZB21Teqt1VvEHx")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xprv9s21ZrQH143K24Mfq5zL5MhWK9hUhhGbd45hLXo2Pq2oqzMMo63oStZzFAzHGBP2UuGCqWLTAPLcMtD5SDKr24z3aiUvKr9bJpdrcLg1y3G")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xpub661MyMwAqRbcEYS8w7XLSVeEsBXy79zSzH1J8vCdxAZningWLdN3zgtU6Q5JXayek4PRsn35jii4veMimro1xefsM58PgBMrvdYre8QyULY")

        with self.assertRaises(ValueError):
            HDWallet.deserialize("xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHL")

    def test_006_bip84(self):
        """Test BIP84 confirmance"""
        mnemonic = "abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about"
        hdw = HDWallet.from_mnemonic(mnemonic)
        assert hdw.serialize_b58(private=True, segwit=True) == \
            "zprvAWgYBBk7JR8Gjrh4UJQ2uJdG1r3WNRRfURiABBE3RvMXYSrRJL62XuezvGdPvG6GFBZduosCc1YP5wixPox7zhZLfiUm8aunE96BBa4Kei5"
        assert hdw.serialize_b58(private=False, segwit=True) == \
            "zpub6jftahH18ngZxLmXaKw3GSZzZsszmt9WqedkyZdezFtWRFBZqsQH5hyUmb4pCEeZGmVfQuP5bedXTB8is6fTv19U1GQRyQUKQGUTzyHACMF"

        # Account 0, root = m/84'/0'/0'
        hdw1 = hdw.get_child_for_path("m/84'/0'/0'")
        assert hdw1.serialize_b58(private=True, segwit=True) == \
            "zprvAdG4iTXWBoARxkkzNpNh8r6Qag3irQB8PzEMkAFeTRXxHpbF9z4QgEvBRmfvqWvGp42t42nvgGpNgYSJA9iefm1yYNZKEm7z6qUWCroSQnE"
        assert hdw1.serialize_b58(private=False, segwit=True) == \
            "zpub6rFR7y4Q2AijBEqTUquhVz398htDFrtymD9xYYfG1m4wAcvPhXNfE3EfH1r1ADqtfSdVCToUG868RvUUkgDKf31mGDtKsAYz2oz2AGutZYs"

        # Account 0, first receiving address = m/84'/0'/0'/0/0
        hdw2 = hdw.get_child_for_path("m/84'/0'/0'/0/0")
        assert hdw2.private_key.to_wif(compressed=True) == \
            "KyZpNDKnfs94vbrwhJneDi77V6jF64PWPF8x5cdJb8ifgg2DUc9d"
        assert hdw2.public_key.to_hex(compressed=True) == \
            "0330d54fd0dd420a6e5f8d3624f5f3482cae350f79d5f0753bf5beef9c2d91af3c"
        assert hdw2.public_key.bech32_address(compressed=True, witness_version=0) == \
            "bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu"

        # Account 0, second receiving address = m/84'/0'/0'/0/1
        hdw3 = hdw.get_child_for_path("m/84'/0'/0'/0/1")
        assert hdw3.private_key.to_wif(compressed=True) == \
            "Kxpf5b8p3qX56DKEe5NqWbNUP9MnqoRFzZwHRtsFqhzuvUJsYZCy"
        assert hdw3.public_key.to_hex(compressed=True) == \
            "03e775fd51f0dfb8cd865d9ff1cca2a158cf651fe997fdc9fee9c1d3b5e995ea77"
        assert hdw3.public_key.bech32_address(compressed=True, witness_version=0) == \
            "bc1qnjg0jd8228aq7egyzacy8cys3knf9xvrerkf9g"

        # Account 0, first change address = m/84'/0'/0'/1/0
        hdw4 = hdw.get_child_for_path("m/84'/0'/0'/1/0")
        assert hdw4.private_key.to_wif(compressed=True) == \
            "KxuoxufJL5csa1Wieb2kp29VNdn92Us8CoaUG3aGtPtcF3AzeXvF"
        assert hdw4.public_key.to_hex(compressed=True) == \
            "03025324888e429ab8e3dbaf1f7802648b9cd01e9b418485c5fa4c1b9b5700e1a6"
        assert hdw4.public_key.bech32_address(compressed=True, witness_version=0) == \
            "bc1q8c6fshw2dlwun7ekn9qwf37cu2rn755upcp6el"

    def test_007_brainwallet(self):
        """Tests brainwallet generation."""
        w = HDWallet.from_brainwallet("crazy horse battery staple", network=BitcoinSegwitMainNet)
        assert w.private_key.to_wif(compressed=True) == "KzJp5B7mDpZ7kMHv67GowQRys9W9Hbaa5Rzj4PCoiyXfTk1fGAvH"

    def test_008_bip38(self):
        """Tests BIP38 key generation (non-ECC mode)"""
        key = PrivateKey.from_wif("5KN7MzqK5wt2TP1fQCYyHBtDrXdJuXbUzm4A9rKAteGu3Qi5CVR")
        bip38key = Bip38PrivateKey(key, "TestingOneTwoThree", compressed=False)
        assert bip38key.base58 == "6PRVWUbkzzsbcVac2qwfssoUJAN1Xhrg6bNk8J7Nzm5H7kxEbn2Nh2ZoGg"
        assert bip38key.private_key("TestingOneTwoThree", compressed=False).to_wif(compressed=False) == "5KN7MzqK5wt2TP1fQCYyHBtDrXdJuXbUzm4A9rKAteGu3Qi5CVR"



'''Base58 encoding

Implementations of Base58 and Base58Check encodings that are compatible
with the bitcoin network.
'''

# This module is based upon base58 snippets found scattered over many bitcoin
# tools written in python. From what I gather the original source is from a
# forum post by Gavin Andresen, so direct your praise to him.
# This module adds shiny packaging and support for python3.

from functools import lru_cache
from hashlib import sha256
from typing import Mapping, Union

__version__ = '2.1.1'

# 58 character alphabet used
BITCOIN_ALPHABET = \
    b'123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
RIPPLE_ALPHABET = b'rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1tuvAxyz'
XRP_ALPHABET = RIPPLE_ALPHABET

# Retro compatibility
alphabet = BITCOIN_ALPHABET


def scrub_input(v: Union[str, bytes]) -> bytes:
    if isinstance(v, str):
        v = v.encode('ascii')

    return v


def b58encode_int(
    i: int, default_one: bool = True, alphabet: bytes = BITCOIN_ALPHABET
) -> bytes:
    """
    Encode an integer using Base58
    """
    if not i and default_one:
        return alphabet[0:1]
    string = b""
    base = len(alphabet)
    while i:
        i, idx = divmod(i, base)
        string = alphabet[idx:idx+1] + string
    return string


def b58encode(
    v: Union[str, bytes], alphabet: bytes = BITCOIN_ALPHABET
) -> bytes:
    """
    Encode a string using Base58
    """
    v = scrub_input(v)

    origlen = len(v)
    v = v.lstrip(b'\0')
    newlen = len(v)

    acc = int.from_bytes(v, byteorder='big')  # first byte is most significant

    result = b58encode_int(acc, default_one=False, alphabet=alphabet)
    return alphabet[0:1] * (origlen - newlen) + result


@lru_cache()
def _get_base58_decode_map(alphabet: bytes,
                           autofix: bool) -> Mapping[int, int]:
    invmap = {char: index for index, char in enumerate(alphabet)}

    if autofix:
        groups = [b'0Oo', b'Il1']
        for group in groups:
            pivots = [c for c in group if c in invmap]
            if len(pivots) == 1:
                for alternative in group:
                    invmap[alternative] = invmap[pivots[0]]

    return invmap


def b58decode_int(
    v: Union[str, bytes], alphabet: bytes = BITCOIN_ALPHABET, *,
    autofix: bool = False
) -> int:
    """
    Decode a Base58 encoded string as an integer
    """
    if b' ' not in alphabet:
        v = v.rstrip()
    v = scrub_input(v)

    map = _get_base58_decode_map(alphabet, autofix=autofix)

    decimal = 0
    base = len(alphabet)
    try:
        for char in v:
            decimal = decimal * base + map[char]
    except KeyError as e:
        raise ValueError(
            "Invalid character {!r}".format(chr(e.args[0]))
        ) from None
    return decimal


def b58decode(
    v: Union[str, bytes], alphabet: bytes = BITCOIN_ALPHABET, *,
    autofix: bool = False
) -> bytes:
    """
    Decode a Base58 encoded string
    """
    v = v.rstrip()
    v = scrub_input(v)

    origlen = len(v)
    v = v.lstrip(alphabet[0:1])
    newlen = len(v)

    acc = b58decode_int(v, alphabet=alphabet, autofix=autofix)

    return acc.to_bytes(origlen - newlen + (acc.bit_length() + 7) // 8, 'big')


def b58encode_check(
    v: Union[str, bytes], alphabet: bytes = BITCOIN_ALPHABET
) -> bytes:
    """
    Encode a string using Base58 with a 4 character checksum
    """
    v = scrub_input(v)

    digest = sha256(sha256(v).digest()).digest()
    return b58encode(v + digest[:4], alphabet=alphabet)


def b58decode_check(
    v: Union[str, bytes], alphabet: bytes = BITCOIN_ALPHABET, *,
    autofix: bool = False
) -> bytes:
    '''Decode and verify the checksum of a Base58 encoded string'''

    result = b58decode(v, alphabet=alphabet, autofix=autofix)
    result, check = result[:-4], result[-4:]
    digest = sha256(sha256(result).digest()).digest()

    if check != digest[:4]:
        raise ValueError("Invalid checksum")

    return result


#
# Copyright (c) 2013 Pavol Rusnak
# Copyright (c) 2017 mruddy
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import hashlib
import hmac
import itertools
import os
import secrets
import unicodedata
from typing import AnyStr, List, TypeVar, Union

_T = TypeVar("_T")
PBKDF2_ROUNDS = 2048


class ConfigurationError(Exception):
    pass


# Refactored code segments from <https://github.com/keis/base58>
def b58encode(v: bytes) -> str:
    alphabet = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"

    p, acc = 1, 0
    for c in reversed(v):
        acc += p * c
        p = p << 8

    string = ""
    while acc:
        acc, idx = divmod(acc, 58)
        string = alphabet[idx : idx + 1] + string
    return string


class Mnemonic(object):
    def __init__(self, language: str = "english"):
        self.language = language
        self.radix = 2048
        d = os.path.join(os.path.dirname(__file__), f"wordlist/{language}.txt")
        if os.path.exists(d) and os.path.isfile(d):
            with open(d, "r", encoding="utf-8") as f:
                self.wordlist = [w.strip() for w in f.readlines()]
            if len(self.wordlist) != self.radix:
                raise ConfigurationError(
                    f"Wordlist should contain {self.radix} words, but it's {len(self.wordlist)} words long instead."
                )
        else:
            raise ConfigurationError("Language not detected")
        # Japanese must be joined by ideographic space
        self.delimiter = "\u3000" if language == "japanese" else " "

    @classmethod
    def list_languages(cls) -> List[str]:
        return [
            f.split(".")[0]
            for f in os.listdir(os.path.join(os.path.dirname(__file__), "wordlist"))
            if f.endswith(".txt")
        ]

    @staticmethod
    def normalize_string(txt: AnyStr) -> str:
        if isinstance(txt, bytes):
            utxt = txt.decode("utf8")
        elif isinstance(txt, str):
            utxt = txt
        else:
            raise TypeError("String value expected")

        return unicodedata.normalize("NFKD", utxt)

    @classmethod
    def detect_language(cls, code: str) -> str:
        """Scan the Mnemonic until the language becomes unambiguous."""
        code = cls.normalize_string(code)
        possible = set(cls(lang) for lang in cls.list_languages())
        for word in code.split():
            possible = set(p for p in possible if word in p.wordlist)
            if not possible:
                raise ConfigurationError(f"Language unrecognized for {word!r}")
        if len(possible) == 1:
            return possible.pop().language
        raise ConfigurationError(
            f"Language ambiguous between {', '.join( p.language for p in possible)}"
        )

    def generate(self, strength: int = 128) -> str:
        """
        Create a new mnemonic using a random generated number as entropy.

        As defined in BIP39, the entropy must be a multiple of 32 bits, and its size must be between 128 and 256 bits.
        Therefore the possible values for `strength` are 128, 160, 192, 224 and 256.

        If not provided, the default entropy length will be set to 128 bits.

        The return is a list of words that encodes the generated entropy.

        :param strength: Number of bytes used as entropy
        :type strength: int
        :return: A randomly generated mnemonic
        :rtype: str
        """
        if strength not in [128, 160, 192, 224, 256]:
            raise ValueError(
                "Invalid strength value. Allowed values are [128, 160, 192, 224, 256]."
            )
        return self.to_mnemonic(secrets.token_bytes(strength // 8))

    # Adapted from <http://tinyurl.com/oxmn476>
    def to_entropy(self, words: Union[List[str], str]) -> bytearray:
        if not isinstance(words, list):
            words = words.split(" ")
        if len(words) not in [12, 15, 18, 21, 24]:
            raise ValueError(
                "Number of words must be one of the following: [12, 15, 18, 21, 24], but it is not (%d)."
                % len(words)
            )
        # Look up all the words in the list and construct the
        # concatenation of the original entropy and the checksum.
        concatLenBits = len(words) * 11
        concatBits = [False] * concatLenBits
        wordindex = 0
        for word in words:
            # Find the words index in the wordlist
            ndx = self.wordlist.index(word)
            if ndx < 0:
                raise LookupError('Unable to find "%s" in word list.' % word)
            # Set the next 11 bits to the value of the index.
            for ii in range(11):
                concatBits[(wordindex * 11) + ii] = (ndx & (1 << (10 - ii))) != 0
            wordindex += 1
        checksumLengthBits = concatLenBits // 33
        entropyLengthBits = concatLenBits - checksumLengthBits
        # Extract original entropy as bytes.
        entropy = bytearray(entropyLengthBits // 8)
        for ii in range(len(entropy)):
            for jj in range(8):
                if concatBits[(ii * 8) + jj]:
                    entropy[ii] |= 1 << (7 - jj)
        # Take the digest of the entropy.
        hashBytes = hashlib.sha256(entropy).digest()
        hashBits = list(
            itertools.chain.from_iterable(
                [c & (1 << (7 - i)) != 0 for i in range(8)] for c in hashBytes
            )
        )
        # Check all the checksum bits.
        for i in range(checksumLengthBits):
            if concatBits[entropyLengthBits + i] != hashBits[i]:
                raise ValueError("Failed checksum.")
        return entropy

    def to_mnemonic(self, data: bytes) -> str:
        if len(data) not in [16, 20, 24, 28, 32]:
            raise ValueError(
                f"Data length should be one of the following: [16, 20, 24, 28, 32], but it is not {len(data)}."
            )
        h = hashlib.sha256(data).hexdigest()
        b = (
            bin(int.from_bytes(data, byteorder="big"))[2:].zfill(len(data) * 8)
            + bin(int(h, 16))[2:].zfill(256)[: len(data) * 8 // 32]
        )
        result = []
        for i in range(len(b) // 11):
            idx = int(b[i * 11 : (i + 1) * 11], 2)
            result.append(self.wordlist[idx])
        return self.delimiter.join(result)

    def check(self, mnemonic: str) -> bool:
        mnemonic_list = self.normalize_string(mnemonic).split(" ")
        # list of valid mnemonic lengths
        if len(mnemonic_list) not in [12, 15, 18, 21, 24]:
            return False
        try:
            idx = map(
                lambda x: bin(self.wordlist.index(x))[2:].zfill(11), mnemonic_list
            )
            b = "".join(idx)
        except ValueError:
            return False
        l = len(b)  # noqa: E741
        d = b[: l // 33 * 32]
        h = b[-l // 33 :]
        nd = int(d, 2).to_bytes(l // 33 * 4, byteorder="big")
        nh = bin(int(hashlib.sha256(nd).hexdigest(), 16))[2:].zfill(256)[: l // 33]
        return h == nh

    def expand_word(self, prefix: str) -> str:
        if prefix in self.wordlist:
            return prefix
        else:
            matches = [word for word in self.wordlist if word.startswith(prefix)]
            if len(matches) == 1:  # matched exactly one word in the wordlist
                return matches[0]
            else:
                # exact match not found.
                # this is not a validation routine, just return the input
                return prefix

    def expand(self, mnemonic: str) -> str:
        return " ".join(map(self.expand_word, mnemonic.split(" ")))

    @classmethod
    def to_seed(cls, mnemonic: str, passphrase: str = "") -> bytes:
        mnemonic = cls.normalize_string(mnemonic)
        passphrase = cls.normalize_string(passphrase)
        passphrase = "mnemonic" + passphrase
        mnemonic_bytes = mnemonic.encode("utf-8")
        passphrase_bytes = passphrase.encode("utf-8")
        stretched = hashlib.pbkdf2_hmac(
            "sha512", mnemonic_bytes, passphrase_bytes, PBKDF2_ROUNDS
        )
        return stretched[:64]

    @staticmethod
    def to_hd_master_key(seed: bytes, testnet: bool = False) -> str:
        if len(seed) != 64:
            raise ValueError("Provided seed should have length of 64")

        # Compute HMAC-SHA512 of seed
        seed = hmac.new(b"Bitcoin seed", seed, digestmod=hashlib.sha512).digest()

        # Serialization format can be found at: https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki#serialization-format
        xprv = b"\x04\x88\xad\xe4"  # Version for private mainnet
        if testnet:
            xprv = b"\x04\x35\x83\x94"  # Version for private testnet
        xprv += b"\x00" * 9  # Depth, parent fingerprint, and child number
        xprv += seed[32:]  # Chain code
        xprv += b"\x00" + seed[:32]  # Master key

        # Double hash using SHA256
        hashed_xprv = hashlib.sha256(xprv).digest()
        hashed_xprv = hashlib.sha256(hashed_xprv).digest()

        # Append 4 bytes of checksum
        xprv += hashed_xprv[:4]

        # Return base58
        return b58encode(xprv)



from hashlib import sha256
import re
import datetime

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from .ripemd160 import ripemd160
from .keccak import Keccak256



def ensure_bytes(data):
    if not isinstance(data, bytes):
        return data.encode('utf-8')
    return data


def ensure_str(data):
    if isinstance(data, bytes):
        return data.decode('utf-8')
    elif not isinstance(data, str):
        raise ValueError("Invalid value for string")
    return data


def hash160(data):
    """Return ripemd160(sha256(data))"""
    return ripemd160(sha256(data).digest())

def is_hex_string(string):
    """Check if the string is only composed of hex characters."""
    pattern = re.compile(r'[A-Fa-f0-9]+')
    if isinstance(string, bytes):
        string = str(string)
    return pattern.match(string) is not None


def long_to_hex(l, size):
    """Encode a long value as a hex string, 0-padding to size.

    Note that size is the size of the resulting hex string. So, for a 32Byte
    long size should be 64 (two hex characters per byte"."""
    f_str = "{0:0%sx}" % size
    return ensure_bytes(f_str.format(l).lower())

def encrypt(raw, passphrase):
    """
    Encrypt text with the passphrase
    @param raw: string Text to encrypt
    @param passphrase: string Passphrase
    @type raw: string
    @type passphrase: string
    @rtype: bytes
    """
    backend = default_backend()
    cipher = Cipher(algorithms.AES(passphrase), modes.CBC(b'\x00'*16), backend=backend)
    encryptor = cipher.encryptor()
    return encryptor.update(raw) + encryptor.finalize()

def decrypt(enc, passphrase):
    """
    Decrypt encrypted text with the passphrase
    @param enc: bytes Text to decrypt
    @param passphrase: string Passphrase
    @type enc: bytes
    @type passphrase: string
    @rtype: bytes
    """
    backend = default_backend()
    cipher = Cipher(algorithms.AES(passphrase), modes.CBC(b'\x00'*16), backend=backend)
    decryptor = cipher.decryptor()
    return decryptor.update(enc) + decryptor.finalize()

def convert_to_utc_timestamp(date_string, format_string="%Y-%m-%dT%H:%M:%SZ"):
    # Create a datetime object from the input string
    # I think we have to check what value the format_string is supposed to be.
    date_object = datetime.datetime.strptime(date_string, format_string)

    # Convert the datetime object to UTC timezone
    utc_date = date_object.astimezone(datetime.timezone.utc)

    # Calculate the timestamp
    timestamp = int(utc_date.timestamp())
    
    return timestamp


def eth_transaction_hash(address: str, nonce: int) -> str:
    """Constructs an Ethereum transaction hash from an address and a transaction number"""
    # Combine the address and nonce as a string
    data_to_hash = address.lower() + format(nonce, 'x')
    
    # Calculate the hash using keccak256
    transaction_hash = Keccak256(bytes.fromhex(data_to_hash)).digest().hex()
    return transaction_hash


class KeyParseError(Exception):
    pass


class ChecksumException(Exception):
    """
    Raised when the generated checksum of an address or public key
    does not match the expected checksum.
    """


class IncompatibleNetworkException(Exception):
    """
    Raised when importing a piece of network data
    into the wrong network.
    """


class InvalidChildException(Exception):
    """
    Raised when an invalid child key is passed.
    """


class WatchOnlyWalletError(Exception):
    """
    Raised when a wallet does not contain a private key when
    performing an operation that requires a public key.
    """

class SegwitError(Exception):
    """
    Raised when a network does not support Segwit (P2WPKH).
    """

class InvalidPathError(Exception):
    """
    Raised when the provided derivation path is invalid.
    """

class NetworkException(Exception):
    """
    Raised when a network request fails.
    """

def incompatible_network_bytes_exception_factory(
        network_name, expected_prefix, given_prefix):
    """A factory function for IncompatibleNetworkException."""
    return IncompatibleNetworkException(
        f"Incorrect network. {network_name} expects a byte prefix of "
        f"{expected_prefix}, but you supplied {given_prefix}")


def unsupported_feature_exception_factory(
        network_name, feature):
    """Another factory function for IncompatibleNetworkException."""
    return IncompatibleNetworkException(
        f"{network_name} does not support feature: {feature}")


import binascii
import hashlib
import web3
from ..network import BitcoinSegwitMainNet
from ..utils.keys import PrivateKey

from ..utxo import UTXO
from ..destination import Destination

from ..utils.base58 import b58decode_check
from ..utils.bech32 import bech32_decode
from ..utils.keccak import to_checksum_address

# Should really use list[] annotation directly but we still support Python 3.8 which does not have such syntax yet.
from typing import List

SIGHASH_ALL = 1

# def address_is_p2pkh(address, network=BitcoinSegwitMainNet):
#     """Checks whether the address is P2PKH"""
#     if "BASE58" not in network.ADDRESS_MODE:
#         return False
#     try:
#         b = b58decode_check(address)
#         return b[0] == network.PUBKEY_ADDRESS
#     except KeyError:
#         return False
#     except ValueError:
#         return False

# def address_is_p2sh(address, network=BitcoinSegwitMainNet):
#     """Checks whether the address is P2SH"""
#     if "BASE58" not in network.ADDRESS_MODE:
#         return False
#     try:
#         b = b58decode_check(address)
#         return b[0] == network.SCRIPT_ADDRESS
#     except KeyError:
#         return False
#     except ValueError:
#         return False

# def address_is_p2wpkh(address, network=BitcoinSegwitMainNet):
#     """Checks whether the address is P2WPKH"""
#     if "BECH32" not in network.ADDRESS_MODE:
#         return False
#     try:
#         b = bech32_decode(network.BECH32_PREFIX, address)
#         return len(b[1]) == 20
#     except KeyError:
#         return False
#     except ValueError:
#         return False
    
# def address_is_p2wsh(address, network=BitcoinSegwitMainNet):
#     """Checks whether the address is P2WSH"""
#     if "BECH32" not in network.ADDRESS_MODE:
#         return False
#     try:
#         b = bech32_decode(network.BECH32_PREFIX, address)
#         return len(b[1]) == 32
#     except KeyError:
#         return False
#     except ValueError:
#         return False

def script_is_p2pkh(script):
    return len(script) == 25 and script[0:3] == b"\x76\xa9\x14" and script[23:25] == b"\x88\xac"

def script_is_p2sh(script):
    return len(script) == 23 and script[0:2] == b"\xa9\x14" and script[24] == b"\x87"

def script_is_p2wpkh(script):
    return len(script) == 22 and script[0:2] == b"\x00\x14"

def script_is_p2wsh(script):
    return len(script) == 34 and script[0:2] == b"\x00\x20"


def int_to_hex(i, min_bytes=1):
    return i.to_bytes(max(min_bytes, (i.bit_length() + 7) // 8), byteorder="little")

def create_varint(value):
    if value < 0xfd:
        return int_to_hex(value, min_bytes=1)
    elif value <= 0xffff:
        return b'\xfd' + int_to_hex(value, min_bytes=2)
    elif value <= 0xffffffff:
        return b'\xfe' + int_to_hex(value, min_bytes=4)
    else:
        return b'\xff' + int_to_hex(value, min_bytes=8)


def create_signatures_legacy(bytes_1, bytes_2_inputs, bytes_3, bytes_4, network):
    """
    Signs the inputs of a legacy transaction. The parts are contained in bytes 1, 2, 3, 4.
    Bytes 2 contains the inputs broken up so that the signature is isolated. It also has
    the script pubkey, the private key, and sighash.
    Note that Segwit transactions use a different signing format (see BIP 143).
    """
    signatures = []
    # Note that only ONE INPUT IS FILLED AT A TIME DURING SIGNING
    for b2i in range(0, len(bytes_2_inputs)):
        b2 = bytes_2_inputs[b2i]
        script_pubkey = b2[3]
        private_key = b2[4]
        sighash = b2[5]
        address = b2[6]
        partial_transaction = bytes_1
        for i in range(0, len(bytes_2_inputs)):
            partial_transaction += bytes_2_inputs[i][0]
            if i == b2i:
                partial_transaction += create_varint(len(script_pubkey)) + script_pubkey
            else:
                partial_transaction += bytes_2_inputs[i][1] # The empty scriptsig
            partial_transaction += bytes_2_inputs[i][2]
        partial_transaction += bytes_3
        partial_transaction += bytes_4
        # And last, the input's sighash must be placed AT THE END of the temporary transaction
        partial_transaction += int_to_hex(sighash, 4)
        hashed_preimage = hashlib.sha256(partial_transaction).digest()

        # Sign it
        p = PrivateKey.from_wif(private_key, network=network)
        pubkey = p.public_key.to_bytes()
        if p.public_key.base58_address(compressed=False) == address:
            pubkey = p.public_key.to_bytes(False)
        der = p.der_sign(hashed_preimage) + bytes([sighash])
        # I would like to mention that this only works if the data being pushed is less than 76 bytes.
        # Otherwise we need to use OP_PUSHDATA<1/2/4>
        script = int_to_hex(len(der)) + der + int_to_hex(len(pubkey)) + pubkey
        signatures.append(create_varint(len(script)) + script)

    # Now that we have all the signatures, we can assemble the signed transaction
    signed_transaction = bytes_1
    for i in range(0, len(bytes_2_inputs)):
        signed_transaction += bytes_2_inputs[i][0]
        signed_transaction += signatures[i]
        signed_transaction += bytes_2_inputs[i][2]
    signed_transaction += bytes_3
    signed_transaction += bytes_4

    return signed_transaction.hex()
    


def create_signatures_segwit(bytes_1, bytes_2_inputs, bytes_3, bytes_4, network):
    """
    Signs the inputs of a segwit transaction. The parts are contained in bytes 1, 2, 3, 4.
    Bytes 2 contains the inputs broken up so that the signature is isolated. It also has
    the script pubkey, the private key, and sighash.
    """
    # The partial transaction to sign is a double SHA256 of the serialization of:
    # 1.  nVersion of the transaction (4-byte little endian)
    # 2.  hashPrevouts (32-byte hash)
    # 3.  hashSequence (32-byte hash)
    # 4.  outpoint (32-byte hash + 4-byte little endian) 
    # 5.  scriptCode of the input (serialized as scripts inside CTxOuts)
    # 6.  value of the output spent by this input (8-byte little endian)
    # 7.  nSequence of the input (4-byte little endian)
    # 8.  hashOutputs (32-byte hash)
    # 9.  nLocktime of the transaction (4-byte little endian)
    # 10. sighash type of the signature (4-byte little endian)
    
    signatures = []
    witness_stack = []
    # Note that only ONE INPUT IS FILLED AT A TIME DURING SIGNING
    for b2i in range(0, len(bytes_2_inputs)):
        b2 = bytes_2_inputs[b2i]
        script_pubkey = b2[3]
        private_key = b2[4]
        sighash = b2[5]
        address = b2[6]
        segwit_payload = hashlib.sha256(b2[7]).digest()

        # Sign it
        p = PrivateKey.from_wif(private_key, network=network)
        pubkey = p.public_key.to_bytes()
        if script_is_p2pkh(script_pubkey):
            if p.public_key.base58_address(compressed=False) == address:
                pubkey = p.public_key.to_bytes(False)
            # Legacy inputs are signed the old way
            der = p.der_sign(hashlib.sha256(segwit_payload).digest()) + bytes([sighash])
            script = int_to_hex(len(der)) + der + int_to_hex(len(pubkey)) + pubkey
            signatures.append(create_varint(len(script)) + script)
            witness_stack.append([])
        elif script_is_p2wpkh(script_pubkey):
            # Place data on the witness stack
            der = p.der_sign(segwit_payload) + bytes([sighash])
            witness_stack.append([der, pubkey])
            signatures.append(b"\x00")
        else:
            raise ValueError("Unsupported script type")
    
    # Now that we have all the signatures, we can assemble the signed transaction
    signed_transaction = bytes_1
    for i in range(0, len(bytes_2_inputs)):
        signed_transaction += bytes_2_inputs[i][0]
        signed_transaction += signatures[i]
        signed_transaction += bytes_2_inputs[i][2]
    signed_transaction += bytes_3

    # Assemble the witness stack, one per input, segwit inputs only
    for w in witness_stack:
        signed_transaction += create_varint(len(w))
        witness_bytes = b""
        for w_elem in w:
            witness_bytes += create_varint(len(w_elem)) + w_elem
        signed_transaction += witness_bytes

    signed_transaction += bytes_4

    return signed_transaction.hex()
    

def create_transaction(inputs: List[UTXO], outputs: List[Destination], rbf=True, network=BitcoinSegwitMainNet, full_nodes=[], **kwargs):
    """
    Creates a signed transaction, given a network an array of UTXOs, and an array of address/amount
    destination tuples.
    Never call this function directly, it does not automatically calculate change or fees. Instead, use Wallet.create_transaction().
    """
    
    # First, construct the raw transacation
    if network.SUPPORTS_EVM:
        try:
            return create_web3_transaction(inputs[0].address(), outputs[0].address(), outputs[0].amount(in_standard_units=False),
                                           inputs[0]._private_key(), full_nodes, kwargs['gas'], kwargs['gasPrice'])
        finally:
            del(inputs)
    else:
        all_legacy = True
        for i in inputs:
            try:
                b58decode_check(i.address())
            except ValueError:
                all_legacy = False
                if not network.SUPPORTS_SEGWIT:
                    raise ValueError("You must use a segwit network to use bech32 inputs")
                break
        tx_bytes_1 = tx_bytes_2 = tx_bytes_3 = tx_bytes_4 = b""
        tx_bytes_1 += int_to_hex(1, 4) # Version 1 transaction
        if network.SUPPORTS_SEGWIT and not all_legacy:
            tx_bytes_1 += b"\x00\x01" # Signal segwit support
        
        # We process the outputs before the inputs so that we can use it for segwit transactions.
        tx_bytes_3 += create_varint(len(outputs))
        tx_bytes_3a = b''
        for o in outputs:
            tx_bytes_3b = b''
            tx_bytes_3b += int_to_hex(o.amount(in_standard_units=False), 8)
            #if network.SUPPORTS_SEGWIT and o.script_pubkey()[0] == 0:
            #    script =  b"\x76\xa9" + o.script_pubkey()[1:] + b"\x88\xac"
            #    tx_bytes_3b += create_varint(len(script)) + script
            #else:
            script = o.script_pubkey()
            tx_bytes_3b += create_varint(len(script))
            tx_bytes_3b += script            
            tx_bytes_3a += tx_bytes_3b
        tx_bytes_3 += tx_bytes_3a

        # Inputs
        tx_bytes_1 += create_varint(len(inputs))
        tx_bytes_2_inputs = []
        for i in inputs:
            input_bytes_1 = input_bytes_2 = input_bytes_3 = input_bytes_4 = b""
            input_bytes_1 += binascii.unhexlify(i.txid().encode())[::-1]
            input_bytes_1 += int_to_hex(i.index(), 4)

            # The transacion cannot be signed until it is fully constructed.
            # To avoid a chicken-and-egg, we set the signature scripts to empty.
            # This is the prescribed behavior by the bitcoin protocol.
            # EDIT I heard it's just the scriptpubkey
            input_bytes_2 = b"\x00" #create_varint(len(i._script_pubkey())) + i._script_pubkey()

            input_bytes_3 = int_to_hex(0xfffffffd if rbf else 0xffffffff, 4) # disables timelocks, see https://bitcointalk.org/index.php?topic=5479540.msg63401889#msg63401889

            segwit_payload = b""
            # It is easier to prepare the Segwit signing data here.
            if network.SUPPORTS_SEGWIT and not all_legacy:
                # nVersion of the transaction (4-byte little endian)
                segwit_payload = int_to_hex(1, 4)
                
                # hashPrevouts (32-byte hash)
                hashPrevouts = b""
                for j in inputs:
                    hashPrevouts += binascii.unhexlify(j.txid().encode())[::-1] + int_to_hex(j.index(), 4)
                segwit_payload += hashlib.sha256(hashlib.sha256(hashPrevouts).digest()).digest()

                # hashSequence (32-byte hash)
                hashSequence = b""
                for j in inputs:
                    hashSequence += input_bytes_3 # The timelock is the same for all inputs.
                segwit_payload += hashlib.sha256(hashlib.sha256(hashSequence).digest()).digest()

                # outpoint (32-byte hash + 4-byte little endian)
                segwit_payload += binascii.unhexlify(i.txid().encode())[::-1] + int_to_hex(i.index(), 4)

                # scriptCode of the input (serialized as scripts inside CTxOuts)
                # note: for p2wpkh this is actually the P2PKH script!!!
                script =  b"\x76\xa9" + i._script_pubkey()[1:] + b"\x88\xac"
                segwit_payload += create_varint(len(script)) + script
                #segwit_payload += create_varint(len(i._script_pubkey())) + i._script_pubkey()

                # value of the output spent by this input (8-byte little endian)
                segwit_payload += int_to_hex(i.amount(in_standard_units=False), 8)

                # nSequence of the input (4-byte little endian)
                segwit_payload += input_bytes_3

                # hashOutputs (32-byte hash)
                segwit_payload += hashlib.sha256(hashlib.sha256(tx_bytes_3a).digest()).digest()

                # nLocktime of the transaction (4-byte little endian)
                segwit_payload += int_to_hex(0, 4)

                # sighash type of the signature (4-byte little endian)
                segwit_payload += int_to_hex(SIGHASH_ALL, 4)


            # If this is a segwit transaction these will need to go into witness data eventually.
            tx_bytes_2_inputs.append([input_bytes_1, input_bytes_2, input_bytes_3, i._script_pubkey(), i._private_key(), SIGHASH_ALL, i.address(), segwit_payload])
        
        # tx_bytes_3 should also contain the witness data

        tx_bytes_4 += int_to_hex(0, 4) # Disable locktime (redundant)

        del(inputs)
        if network.SUPPORTS_SEGWIT and not all_legacy:
            return create_signatures_segwit(tx_bytes_1, tx_bytes_2_inputs, tx_bytes_3, tx_bytes_4, network)
        else:
            return create_signatures_legacy(tx_bytes_1, tx_bytes_2_inputs, tx_bytes_3, tx_bytes_4, network)


def create_web3_transaction(a_from, a_to, amount, private_key, fullnodes, gas, gasPrice):
    sender_address = a_from
    receiver_address = a_to
    # All amounts are in WEI not Ether

    # Check the nonce for the sender address
    for node in fullnodes:
        try:
            w3 = web3.Web3(web3.HTTPProvider(node['url']))
            nonce = w3.eth.getTransactionCount(to_checksum_address(sender_address))

            # Build the transaction dictionary
            transaction = {
                'nonce': nonce,
                'to': to_checksum_address(receiver_address),
                'value': w3.toWei(amount, 'ether'),  # Sending 1 ether, adjust as needed
                'gas': gas,#21000,  # Gas limit
                'gasPrice': w3.toWei(gasPrice, 'gwei'),  # Gas price in Gwei, adjust as needed
                'chainId': 1,  # Mainnet, change to 3 for Ropsten, 4 for Rinkeby, etc.
            }

            # Sign the transaction
            return w3.eth.account.signTransaction(transaction, binascii.unhexlify(private_key[2:].encode()))
        except Exception:
            pass
    raise RuntimeError("Cannot sign web3 transaction (try specifying different nodes)")



from .blockchain_info import BlockchainInfoAddress
from .blockcypher import BlockcypherAddress
from .blockstream import BlockstreamAddress
from .btcdotcom import BTCDotComAddress
from .esplora import EsploraAddress
from .fullnode import BitcoinRPCClient
from .mempoolspace import MempoolSpaceAddress
from ...generated import wallet_pb2
from ...errors import NetworkException
from ...nodes.btc import btc_nodes, btc_esplora_nodes

class BitcoinAddress:
    """ Load balancer for all BTC address providers provided to an instance of this class,
        using the round robin scheduling algorithm.
    """

    def __init__(self, addresses, providers: bytes = b'\xff\xff', max_cycles=100,
                 transactions=None, **kwargs):
        provider_bitmask = int.from_bytes(providers, 'big')
        self.provider_list = []
        self.current_index = 0
        self.addresses = addresses
        self.max_cycles = max_cycles
        fullnode_endpoints = kwargs.get('fullnode_endpoints')
        esplora_endpoints = kwargs.get('esplora_endpoints')
        blockcypher_tokens = kwargs.get('blockcypher_tokens')


        # Set everything to an empty list so that providers do not immediately start fetching
        # transactions and to avoid exceptions in loops later in this method.
        if not transactions:
            transactions = []
        if not esplora_endpoints:
            esplora_endpoints = [] + btc_esplora_nodes
        if not fullnode_endpoints:
            fullnode_endpoints = [] + btc_nodes

        self.transactions = transactions

        #if provider_bitmask & 1 << wallet_pb2.BTC_BLOCKCHAIN_INFO + 1:
        #    self.provider_list.append(BlockchainInfoAddress(addresses, transactions=transactions))
        if provider_bitmask & 1 << wallet_pb2.BTC_BLOCKCYPHER + 1:
            tokens = blockcypher_tokens
            if not tokens:
                tokens = []
            for token in tokens:
                self.provider_list.append(BlockcypherAddress(addresses, transactions=transactions, api_key=token))
            self.provider_list.append(BlockcypherAddress(addresses, transactions=transactions)) # No token (free) version
        if provider_bitmask & 1 << wallet_pb2.BTC_BLOCKSTREAM + 1:
            self.provider_list.append(BlockstreamAddress(addresses, transactions=transactions))
        if provider_bitmask & 1 << wallet_pb2.BTC_BTCDOTCOM + 1:
            self.provider_list.append(BTCDotComAddress(addresses, transactions=transactions))
        if provider_bitmask & 1 << wallet_pb2.BTC_ESPLORA + 1:
            for endpoint in esplora_endpoints:
                self.provider_list.append(EsploraAddress(addresses, transactions=transactions, **endpoint))
        if provider_bitmask & 1 << wallet_pb2.BTC_FULLNODE + 1:
            for endpoint in fullnode_endpoints:
                self.provider_list.append(BitcoinRPCClient(addresses, transactions=transactions, **endpoint))
        if provider_bitmask & 1 << wallet_pb2.BTC_MEMPOOLSPACE + 1:
            self.provider_list.append(MempoolSpaceAddress(addresses, transactions=transactions))

    def sync(self): 
       for provider in self.provider_list:
            try:
                provider.sync()
            except NetworkException:
                pass

    def get_balance(self):
        """
        Retrieves the balance of the Bitcoin address.

        Returns:
            float: The balance of the Bitcoin address in BTC.

        Raises:
            Exception: If the API request fails or the address balance cannot be retrieved.
        """
        utxos = self.get_utxos()
        total_balance = 0
        confirmed_balance = 0
        for utxo in utxos:
            total_balance += utxo.amount
            if utxo.confirmed:
                confirmed_balance += utxo.amount
        return total_balance, confirmed_balance
        
    def get_utxos(self):
        # Transactions are generated in reverse order
        utxos = []
        for i in range(len(self.transactions)-1, -1, -1):
            for out in self.transactions[i].btclike_transaction.outputs:
                if out.spent:
                    continue
                if out.address in self.addresses:
                    utxo = wallet_pb2.UTXO()
                    utxo.address = out.address
                    utxo.txid = self.transactions[i].txid
                    utxo.index = out.index
                    utxo.amount = out.amount
                    utxo.height = self.transactions[i].height
                    utxo.confirmed = self.transactions[i].confirmed
                    utxos.append(utxo)
        return utxos


    def advance_to_next_provider(self):
        if not self.provider_list:
            return
        
        newindex = (self.current_index + 1) % len(self.provider_list)
        self.provider_list[newindex].transactions = self.provider_list[self.current_index].transactions
        self.current_index = newindex
    
    def get_transaction_history(self):
        for address in self.addresses:
            txs = []
            ntransactions = -1  # Set to invalid value for the first iteration
            cycle = 1
            while ntransactions != len(self.transactions):
                if cycle > self.max_cycles:
                    raise NetworkException(f"None of the address providers are working after {self.max_cycles} tries")
                self.provider_list[self.current_index].transactions = txs
                self.provider_list[self.current_index].addresses = [address]
                try:
                    self.provider_list[self.current_index].get_transaction_history()
                    ntransactions = len(self.transactions)
                    txs = self.provider_list[self.current_index].transactions
                    break
                except NetworkException:
                    txs = self.provider_list[self.current_index].transactions
                    self.advance_to_next_provider()
                    cycle += 1
            self.transactions.extend(txs)
        return self.transactions


import requests
import time

from ...errors import NetworkException
from ...generated import wallet_pb2

class BlockchainInfoAddress:
    """
    A class representing a Bitcoin address.

    This class allows you to retrieve the balance and transaction history of a Bitcoin address using the Blockchain.info API.

    The rate limits are 1 request per 10 seconds.

    THIS CLASS IS NOT RECOMMENDED for fetching transaction lists, because there's no way to get the transaction ID of inputs,
    and also because it does not return transaction vsize which is the measurement unit used by segwit blockchains.
    And speaking of blockchains, this backend only supports the Bitcoin mainnet.

    As a side effect of this, it is impossible to determine which transactions are RBFs and/or double-spends and exclude them
    accordingly. This means total balance may be incorrect. However, confirmed balance will still be correct in any case.

    Args:
        address (str): The human-readable Bitcoin address.

    Attributes:
        address (str): The human-readable Bitcoin address.
        transaction_history (list): The cached list of transactions.
        height (int): The last known block height.

    Methods:
        get_balance(): Retrieves the total and confirmed balances of the Bitcoin address.
        get_utxos(): Retrieves the UTXO set for this address
        get_block_height(): Retrieves the current block height.
        get_transaction_history(): Retrieves the transaction history of the Bitcoin address.

    Raises:
        Exception: If the API request fails or the address balance/transaction history cannot be retrieved.
    """
    def __init__(self, addresses, request_interval=(1,10), transactions=None):
        """
        Initializes an instance of the Address class.

        Args:
            addresses (list): A list of human-readable Bitcoin addresses.
            request_interval (tuple): A pair of integers indicating the number of requests allowed during
                a particular amount of seconds. Set to (0,N) for no rate limiting, where N>0.
        """
        self.addresses = addresses
        self.requests, self.interval_sec = request_interval
        if transactions is not None and isinstance(transactions, list):
            self.transactions = transactions
        else:
            self.transactions = []

    def sync(self):
        self.transactions = [*self._get_transaction_history()]
        self.height = self.get_block_height()

    def _clean_tx(self, element):
        new_element = wallet_pb2.Transaction()
        new_element.txid = element['hash']
        if 'block_height' in element.keys() and element['block_height'] is not None:
            new_element.confirmed = True
            new_element.height = element['block_height']
        else:
            new_element.confirmed = False

        if new_element.confirmed:
            new_element.timestamp = element['time']

        for vin in element['inputs']:
            txinput = new_element.btclike_transaction.inputs.add()
            # Blockchain.info is crazy!
            # It has no input txid, only the address
            # So for now, we will substitute the address for the txid,
            # and once we get a sane API, we can fill in the txid properly.
            # (It also only supports Bitcoin, so there's that.)
            txinput.index = vin['prev_out']['n']
            txinput.amount = int(vin['prev_out']['value'] * 1e8)
        
        for vout in element['out']:
            txoutput = new_element.btclike_transaction.outputs.add()
            txoutput.amount = int(vout['value'] * 1e8)
            txoutput.index = vout['n']
            if 'addr' in vout.keys():
                txoutput.address = vout['addr']
            txoutput.spent = vout['spent']
        
        # Now we must calculate the total fee
        total_inputs = sum([a.amount for a in new_element.btclike_transaction.inputs])
        total_outputs = sum([a.amount for a in new_element.btclike_transaction.outputs])
        new_element.total_fee = total_inputs - total_outputs

        # Blockchain.info API does not support vbytes. It only returns bytes.
        new_element.btclike_transaction.fee = int(new_element.total_fee // element['size'])
        new_element.fee_metric = wallet_pb2.BYTE
        
        return new_element

    def get_balance(self):
        """
        Retrieves the balance of the Bitcoin address.

        Returns:
            float: The balance of the Bitcoin address in BTC.

        Raises:
            Exception: If the API request fails or the address balance cannot be retrieved.
        """
        utxos = self.get_utxos()
        total_balance = 0
        confirmed_balance = 0
        for utxo in utxos:
            total_balance += utxo.amount
            if utxo.confirmed:
                confirmed_balance += utxo.amount
        return total_balance, confirmed_balance
        
    def get_utxos(self):
        # Transactions are generated in reverse order
        utxos = []
        for i in range(len(self.transactions)-1, -1, -1):
            for out in self.transactions[i].outputs:
                if out.spent:
                    continue
                if out.address in self.addresses:
                    utxo = wallet_pb2.UTXO()
                    utxo.address = out.address
                    utxo.txid = self.transactions[i].txid
                    utxo.index = out.index
                    utxo.amount = out.amount
                    utxo.height = self.transactions[i].height
                    utxo.confirmed = self.transactions[i].confirmed
                    utxos.append(utxo)
        return utxos
        
        # Now we would've gone through the UTXOs *again* to eliminate the RBF-replaced UTXOs.
        # The issue we are going to have is, *any* of the inputs could be the identical
        # one, because it only takes one input to perform an RBF.

        # Unfortunately we CANNOT filter Blockchain.info UTXOs because they do not give us
        # the txids of inputs, and even the balance endpoint in the Blockchain Data API
        # returns the wrong total balance using exactly the same calculation.
        # Confirmed balances are not affected and still correct.


    def get_block_height(self):
        # Get the current block height now:
        url = "https://blockchain.info/latestblock"
        for attempt in range(3, -1, -1):
            if attempt == 0:
                raise NetworkException("Network request failure")
            try:
                response = requests.get(url, timeout=60)
                break
            except requests.RequestException:
                pass
            except requests.exceptions.JSONDecodeError:
                pass
        if response.status_code == 200:
            return response.json()["height"]
        else:
            raise NetworkException("Cannot get block height")


    def get_transaction_history(self):
        """
        Retrieves the transaction history of the Bitcoin address from cached data augmented with network data.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        self.height = self.get_block_height()
        if len(self.transactions) == 0:
            self.transactions = [*self._get_transaction_history()]
        else:
            # First element is the most recent transactions
            txhash = self.transactions[0].txid
            txs = [*self._get_transaction_history(txhash)]
            txs.extend(self.transactions)
            self.transactions = txs
                    
        return self.transactions

    def _get_transaction_history(self, txhash=None):
        """
        Retrieves the transaction history of the Bitcoin address. (internal method that makes the network query)

        Parameters:
            txhash (str): Get all transactions before (and not including) txhash.
                Defaults to None, which disables this behavior.

        Returns:
            list: A list of dictionaries representing the transaction history.

        Raises:
            Exception: If the API request fails or the transaction history cannot be retrieved.
        """
        for address in self.addresses:
            interval = 50
            offset = 0

            url = f"https://blockchain.info/rawaddr/{address}?limit={interval}&offset={offset}"
            for attempt in range(3, -1, -1):
                if attempt == 0:
                    raise NetworkException("Network request failure")
                try:
                    response = requests.get(url, timeout=60)
                    if response.status_code == 200:
                        data = response.json()
                        break
                    else:
                        raise NetworkException("Failed to retrieve transaction history")
                except requests.RequestException:
                    pass

                # The rate limit is 1 request every 10 seconds, so we will amortize the speed bump by sleeping every 200 milliseconds.
                # Since we have max 50 transactions, total execution time will be at least 10 seconds incuding the sleep time and user
                # code execution time, and if there are less than 50 transactions, we are finished fetching transactions anyway.
                for tx in data["txs"]:
                    time.sleep(self.interval_sec/(self.requests*len(data["txs"])))
                    if txhash and tx["hash"] == txhash:
                        return
                    yield self._clean_tx(tx)
                n_tx = data["n_tx"]
                offset += min(interval, n_tx)
            else:
                raise NetworkException("Failed to retrieve transaction history")
            
            while offset < n_tx:
                # WARNING: RATE LIMIT IS 1 REQUEST PER 10 SECONDS.
                url = f"https://blockchain.info/rawaddr/{address}?limit={interval}&offset={offset}"
                for attempt in range(3, -1, -1):
                    if attempt == 0:
                        raise NetworkException("Network request failure")
                    try:
                        response = requests.get(url, timeout=60)
                        break
                    except requests.RequestException:
                        pass

                if response.status_code == 200:
                    data = response.json()
                    for tx in data["txs"]:
                        time.sleep(self.interval_sec/(self.requests*len(data["txs"])))
                        if txhash and tx["hash"] == txhash:
                            return
                        yield self._clean_tx(tx)
                    offset += interval
                else:
                    raise NetworkException("Failed to retrieve transaction history")



import requests
from ...errors import NetworkException

async def broadcast_transaction_btc_blockchair(raw_transaction_hex):
    api_url = "https://api.blockchair.com/bitcoin/push/transaction"

    payload = {"data": raw_transaction_hex}

    try:
        response = requests.post(api_url, json=payload, timeout=30)
    except Exception as e:
        raise NetworkException("Connection error while broadcasting transaction: {}".format(str(e)))
    
    result = response.json()

    if response.status_code >= 300:
        raise NetworkException(f"Failed to broadcast Bitcoin transaction using Blockchair API: {result.get('message')}")



import requests
from ...errors import NetworkException

async def broadcast_transaction_eth_etherscan(raw_transaction_hex, api_key):
    api_url = f"https://api.etherscan.io/api"
    payload = {
        "module": "proxy",
        "action": "eth_sendRawTransaction",
        "hex": raw_transaction_hex,
        "apikey": api_key,
    }

    try:
        response = requests.get(api_url, params=payload, timeout=30)
    except Exception as e:
        raise NetworkException("Connection error while broadcasting transaction: {}".format(str(e)))
    
    result = response.json()

    if response.status_code >= 300 and result.get("status") == "1":
        raise NetworkException(f"Failed to broadcast Ethereum transaction using Etherscan: {result.get('message')}")



# Tip: Use https://www.comparenodes.com/library/public-endpoints/ to find public RPC endpoints
btctest_nodes = [
    {
        'url': 'https://bitcoin-testnet-archive.allthatnode.com/'
    },
    {
        'url': 'https://bitcoin-testnet.drpc.org/'
    },
]

btctest_esplora_nodes = [
    {
        'url': 'https://mempool.ninja/testnet/api'
    },
    {
        'url': 'https://www.lohkoketju.fi/testnet/api'
    },
]


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: wallet.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cwallet.proto\"6\n\x07RPCNode\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0c\n\x04user\x18\x02 \x01(\t\x12\x10\n\x08password\x18\x03 \x01(\t\"\xe2\x02\n\x06Wallet\x12\x19\n\x07network\x18\x01 \x01(\x0e\x32\x08.Network\x12\x1d\n\x15\x65ncrypted_seed_phrase\x18\x02 \x01(\x0c\x12\x0e\n\x06height\x18\x03 \x01(\x07\x12\x19\n\x11receive_gap_limit\x18\x04 \x01(\x07\x12\x18\n\x10\x63hange_gap_limit\x18\x05 \x01(\x07\x12\x1b\n\taddresses\x18\x06 \x03(\x0b\x32\x08.Address\x12\"\n\x0ctransactions\x18\x07 \x03(\x0b\x32\x0c.Transaction\x12\x18\n\x10\x63rypto_providers\x18\x08 \x01(\r\x12$\n\x12\x66ullnode_endpoints\x18\t \x03(\x0b\x32\x08.RPCNode\x12#\n\x11\x65splora_endpoints\x18\n \x03(\x0b\x32\x08.RPCNode\x12\x1a\n\x12\x62lockcypher_tokens\x18\x0b \x03(\t\x12\x17\n\x0f\x64\x65rivation_path\x18\x0c \x01(\t\";\n\x07\x41\x64\x64ress\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x0e\n\x06pubkey\x18\x02 \x01(\t\x12\x0f\n\x07privkey\x18\x03 \x01(\t\"a\n\x12\x42TCLikeTransaction\x12\x0b\n\x03\x66\x65\x65\x18\x01 \x01(\x04\x12\x1d\n\x06inputs\x18\x02 \x03(\x0b\x32\r.BTCLikeInput\x12\x1f\n\x07outputs\x18\x03 \x03(\x0b\x32\x0e.BTCLikeOutput\"Q\n\x0c\x42TCLikeInput\x12\x0c\n\x04txid\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\r\x12\x0e\n\x06\x61mount\x18\x03 \x01(\x04\x12\x14\n\x0cwitness_data\x18\x04 \x03(\x0c\"N\n\rBTCLikeOutput\x12\x0e\n\x06\x61mount\x18\x01 \x01(\x04\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\r\n\x05index\x18\x03 \x01(\x04\x12\r\n\x05spent\x18\x04 \x01(\x08\"]\n\x12\x45THLikeTransaction\x12\x0e\n\x06txfrom\x18\x01 \x01(\t\x12\x0c\n\x04txto\x18\x02 \x01(\t\x12\x0e\n\x06\x61mount\x18\x03 \x01(\x04\x12\x0b\n\x03gas\x18\x04 \x01(\x04\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\"\xe8\x01\n\x0bTransaction\x12\x0c\n\x04txid\x18\x01 \x01(\t\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\x12\x11\n\tconfirmed\x18\x03 \x01(\x08\x12\x0e\n\x06height\x18\x04 \x01(\x04\x12\x11\n\ttotal_fee\x18\x05 \x01(\x04\x12\x1e\n\nfee_metric\x18\x06 \x01(\x0e\x32\n.FeeMetric\x12\x30\n\x13\x62tclike_transaction\x18\x0f \x01(\x0b\x32\x13.BTCLikeTransaction\x12\x30\n\x13\x65thlike_transaction\x18\x0e \x01(\x0b\x32\x13.ETHLikeTransaction\"{\n\x04UTXO\x12\x0e\n\x06\x61mount\x18\x01 \x01(\x04\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\x0c\n\x04txid\x18\x03 \x01(\t\x12\r\n\x05index\x18\x04 \x01(\x04\x12\x12\n\nprivatekey\x18\x05 \x01(\t\x12\x11\n\tconfirmed\x18\x06 \x01(\x08\x12\x0e\n\x06height\x18\x07 \x01(\x04*\x8b\x04\n\x07Network\x12\x1a\n\x16\x42ITCOIN_SEGWIT_MAINNET\x10\x00\x12\x13\n\x0f\x42ITCOIN_MAINNET\x10\x01\x12\x1a\n\x16\x42ITCOIN_SEGWIT_TESTNET\x10\x02\x12\x13\n\x0f\x42ITCOIN_TESTNET\x10\x03\x12\x1b\n\x17LITECOIN_SEGWIT_MAINNET\x10\x04\x12\x14\n\x10LITECOIN_MAINNET\x10\x05\x12\x1f\n\x1bLITECOIN_BTC_SEGWIT_MAINNET\x10\x06\x12\x18\n\x14LITECOIN_BTC_MAINNET\x10\x07\x12\x1b\n\x17LITECOIN_SEGWIT_TESTNET\x10\x08\x12\x14\n\x10LITECOIN_TESTNET\x10\t\x12\x14\n\x10\x45THEREUM_MAINNET\x10\n\x12\x14\n\x10\x44OGECOIN_MAINNET\x10\x0b\x12\x18\n\x14\x44OGECOIN_BTC_MAINNET\x10\x0c\x12\x14\n\x10\x44OGECOIN_TESTNET\x10\r\x12\x10\n\x0c\x44\x41SH_MAINNET\x10\x0e\x12\x19\n\x15\x44\x41SH_INVERTED_MAINNET\x10\x0f\x12\x14\n\x10\x44\x41SH_BTC_MAINNET\x10\x10\x12\x10\n\x0c\x44\x41SH_TESTNET\x10\x11\x12\x19\n\x15\x44\x41SH_INVERTED_TESTNET\x10\x12\x12\x18\n\x14\x42ITCOIN_CASH_MAINNET\x10\x13\x12\x17\n\x13\x42LOCKCYPHER_TESTNET\x10\x14*\"\n\x0b\x42\x43YProvider\x12\x13\n\x0f\x42\x43Y_BLOCKCYPHER\x10\x00*\x9c\x01\n\x0b\x42TCProvider\x12\x13\n\x0f\x42TC_BLOCKCYPHER\x10\x00\x12\x13\n\x0f\x42TC_BLOCKSTREAM\x10\x01\x12\x14\n\x10\x42TC_MEMPOOLSPACE\x10\x02\x12\x11\n\rBTC_BTCDOTCOM\x10\x03\x12\x17\n\x13\x42TC_BLOCKCHAIN_INFO\x10\x04\x12\x0f\n\x0b\x42TC_ESPLORA\x10\x05\x12\x10\n\x0c\x42TC_FULLNODE\x10\x06*\x88\x01\n\x0f\x42TCTestProvider\x12\x17\n\x13\x42TCTEST_BLOCKCYPHER\x10\x00\x12\x17\n\x13\x42TCTEST_BLOCKSTREAM\x10\x01\x12\x18\n\x14\x42TCTEST_MEMPOOLSPACE\x10\x02\x12\x13\n\x0f\x42TCTEST_ESPLORA\x10\x03\x12\x14\n\x10\x42TCTEST_FULLNODE\x10\x04*7\n\x0c\x44\x41SHProvider\x12\x14\n\x10\x44\x41SH_BLOCKCYPHER\x10\x00\x12\x11\n\rDASH_FULLNODE\x10\x01*8\n\x0c\x44OGEProvider\x12\x14\n\x10\x44OGE_BLOCKCYPHER\x10\x00\x12\x12\n\x0e\x44OGE_DOGECHAIN\x10\x01*\x1f\n\x0b\x45THProvider\x12\x10\n\x0c\x45TH_FULLNODE\x10\x00*4\n\x0bLTCProvider\x12\x13\n\x0fLTC_BLOCKCYPHER\x10\x00\x12\x10\n\x0cLTC_FULLNODE\x10\x01*)\n\tFeeMetric\x12\x08\n\x04\x42YTE\x10\x00\x12\t\n\x05VBYTE\x10\x01\x12\x07\n\x03WEI\x10\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'wallet_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_NETWORK']._serialized_start=1208
  _globals['_NETWORK']._serialized_end=1731
  _globals['_BCYPROVIDER']._serialized_start=1733
  _globals['_BCYPROVIDER']._serialized_end=1767
  _globals['_BTCPROVIDER']._serialized_start=1770
  _globals['_BTCPROVIDER']._serialized_end=1926
  _globals['_BTCTESTPROVIDER']._serialized_start=1929
  _globals['_BTCTESTPROVIDER']._serialized_end=2065
  _globals['_DASHPROVIDER']._serialized_start=2067
  _globals['_DASHPROVIDER']._serialized_end=2122
  _globals['_DOGEPROVIDER']._serialized_start=2124
  _globals['_DOGEPROVIDER']._serialized_end=2180
  _globals['_ETHPROVIDER']._serialized_start=2182
  _globals['_ETHPROVIDER']._serialized_end=2213
  _globals['_LTCPROVIDER']._serialized_start=2215
  _globals['_LTCPROVIDER']._serialized_end=2267
  _globals['_FEEMETRIC']._serialized_start=2269
  _globals['_FEEMETRIC']._serialized_end=2310
  _globals['_RPCNODE']._serialized_start=16
  _globals['_RPCNODE']._serialized_end=70
  _globals['_WALLET']._serialized_start=73
  _globals['_WALLET']._serialized_end=427
  _globals['_ADDRESS']._serialized_start=429
  _globals['_ADDRESS']._serialized_end=488
  _globals['_BTCLIKETRANSACTION']._serialized_start=490
  _globals['_BTCLIKETRANSACTION']._serialized_end=587
  _globals['_BTCLIKEINPUT']._serialized_start=589
  _globals['_BTCLIKEINPUT']._serialized_end=670
  _globals['_BTCLIKEOUTPUT']._serialized_start=672
  _globals['_BTCLIKEOUTPUT']._serialized_end=750
  _globals['_ETHLIKETRANSACTION']._serialized_start=752
  _globals['_ETHLIKETRANSACTION']._serialized_end=845
  _globals['_TRANSACTION']._serialized_start=848
  _globals['_TRANSACTION']._serialized_end=1080
  _globals['_UTXO']._serialized_start=1082
  _globals['_UTXO']._serialized_end=1205
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# flake8: noqa: C0301

"""Tests for using the Wallet class."""


import unittest
from zpywallet.generated import wallet_pb2
from zpywallet import Wallet
from zpywallet.address.btc import BitcoinAddress
from zpywallet.destination import Destination
from zpywallet.network import BitcoinSegwitMainNet
from zpywallet.transactions.encode import create_transaction
from zpywallet.utxo import UTXO
from zpywallet.nodes.btc import btc_nodes

class TestWallet(unittest.TestCase):
    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_create_wallet(self):
        wallet = Wallet(BitcoinSegwitMainNet, None, "zpywallet", receive_gap_limit=1)
        wallet_1 = Wallet(BitcoinSegwitMainNet, "abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon cactus",
                          "zpywallet", receive_gap_limit=1)
        exported_wallet = wallet_1.serialize()
        wallet_2 = Wallet.deserialize(exported_wallet, "zpywallet")
        

    def test_001_use_the_wallet(self):
        """Test using the wallet."""
        wallet = Wallet(BitcoinSegwitMainNet, "abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon cactus",
                          "zpywallet", receive_gap_limit=1)
        wallet.get_transaction_history()
        wallet.get_utxos(only_unspent=False)
        wallet.get_utxos(only_unspent=True)
        wallet.get_balance()
        wallet.addresses()
        wallet.random_address()
        wallet.private_keys("zpywallet")
        # This seems to work off and on inside Pytest/TOX
        #with self.assertRaises(ValueError):
        #    wallet.private_keys("wrongpassword")
        wallet._add_stock_nodes()
        wallet.broadcast_transaction(b"010000000113f3b5446192eeb85c0f7ca64c12196c26314d27b5ab6a976d560cb134ffa82f020000006b483045022100deb556a5c301696f5def1888e54e8c1227caeb3ed4ae7b1c665af79174f50ab702201a59bba811554bd66fa003606ba852bf17699f38c6b1c3f8628f36c9d193a081012102245a4ecd8ad47f171f90c6d4a4f929052814d844e4d8c112bbf799aedc1b8555ffffffff061953cb010000000017a9145bf7f29863984cd9c8ff321a6d93d6344c7c055287102cf604000000001976a9148e09dee91c997fd306030ad7a1c46f17dcd51fdd88ace00f97000000000017a914ecd0679169813020faa97b418be8e79098c1cd55871595c1000000000017a914b61a3c9897d3d539aee6ad7327dfbd2192103e6287a4b62100000000001976a9144954c267b47bd1f864e627ca5b9c82b1fde8966188ac9e41a409000000001976a9141199950f4896dbbbd5ddc833092f5458b5f3154888ac00000000")
        saved_utxos = [b'\x08\xc0\x84=\x12"16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM\x1a@dce83bbde7aba21c8994f9176e827a6a6ce28b4e4121d9090d8fe0b846b74034 \x010\x018\xc6\x82\t',
                       b'\x08\xa8U\x12"16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM\x1a@2de38a49f0079d0aaa8a0b9cfec71b1af935752b609eee0dc1eae56b2162a7e2 \x010\x018\x88\xef\x11',
                       b'\x08\x90N\x12"16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM\x1a@f6206a176b02b2d333cdfaa380251e423e5650ba52ac4f4e170672d2a2263b09 \x040\x018\x91\xf2\x1e',
                       b'\x08\x90N\x12"16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM\x1a@f6206a176b02b2d333cdfaa380251e423e5650ba52ac4f4e170672d2a2263b09 \x050\x018\x91\xf2\x1e',
                       b'\x08\x90N\x12"16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM\x1a@f6206a176b02b2d333cdfaa380251e423e5650ba52ac4f4e170672d2a2263b09 \x060\x018\x91\xf2\x1e',
                       b'\x08\x90N\x12"16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM\x1a@f6206a176b02b2d333cdfaa380251e423e5650ba52ac4f4e170672d2a2263b09 \x080\x018\x91\xf2\x1e',
                       b'\x08\x9e\x9b\x01\x12"16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM\x1a@83d0aaf8e3bd5a240eba1061c572024750c38c847074f7717ef8467b3c4e38930\x018\xe0\xc5!',
                       b'\x08\xa4!\x12"16QaFeudRUt8NYy2yzjm3BMvG4xBbAsBFM\x1a@ca9ae53f5a8bf29825af1438fcde86c4437668f5ed35ec32b9fc4e5a0a42da680\x018\xe1\xc5!']
        utxos = []
        for u in saved_utxos:
            utxo = wallet_pb2.UTXO()
            utxo.ParseFromString(u)
            utxos.append(utxo)
        
        wallet._to_human_friendly_utxo(utxos, [])


from .blockcypher import BlockcypherAddress
from .fullnode import DashRPCClient
from ...generated import wallet_pb2
from ...errors import NetworkException
from ...nodes.dash import dash_nodes

class DashAddress:
    """ Load balancer for all DASH address providers provided to an instance of this class,
        using the round robin scheduling algorithm.
    """

    def __init__(self, addresses, providers: bytes = b'\xff\xff', max_cycles=100,
                 transactions=None, **kwargs):
        provider_bitmask = int.from_bytes(providers, 'big')
        self.provider_list = []
        self.current_index = 0
        self.addresses = addresses
        self.max_cycles = max_cycles
        fullnode_endpoints = kwargs.get('fullnode_endpoints')
        blockcypher_tokens = kwargs.get('blockcypher_tokens')

        # Set everything to an empty list so that providers do not immediately start fetching
        # transactions and to avoid exceptions in loops later in this method.
        if not transactions:
            transactions = []
        if not fullnode_endpoints:
            fullnode_endpoints = [] + dash_nodes

        self.transactions = transactions

        if provider_bitmask & 1 << wallet_pb2.DASH_BLOCKCYPHER + 1:
            tokens = blockcypher_tokens
            if not tokens:
                tokens = []
            for token in tokens:
                self.provider_list.append(BlockcypherAddress(addresses, transactions=transactions, api_key=token))
            self.provider_list.append(BlockcypherAddress(addresses, transactions=transactions)) # No token (free) version
        if provider_bitmask & 1 << wallet_pb2.DASH_FULLNODE + 1:
            for endpoint in fullnode_endpoints:
                self.provider_list.append(DashRPCClient(addresses, transactions=transactions, **endpoint))

    def sync(self): 
       for provider in self.provider_list:
            try:
                provider.sync()
            except NetworkException:
                pass

    def get_balance(self):
        """
        Retrieves the balance of the Dash address.

        Returns:
            float: The balance of the Dash address in DASH.

        Raises:
            Exception: If the API request fails or the address balance cannot be retrieved.
        """
        utxos = self.get_utxos()
        total_balance = 0
        confirmed_balance = 0
        for utxo in utxos:
            total_balance += utxo.amount
            if utxo.confirmed:
                confirmed_balance += utxo.amount
        return total_balance, confirmed_balance
        
    def get_utxos(self):
        # Transactions are generated in reverse order
        utxos = []
        for i in range(len(self.transactions)-1, -1, -1):
            for out in self.transactions[i].btclike_transaction.outputs:
                if out.spent:
                    continue
                if out.address in self.addresses:
                    utxo = wallet_pb2.UTXO()
                    utxo.address = out.address
                    utxo.txid = self.transactions[i].txid
                    utxo.index = out.index
                    utxo.amount = out.amount
                    utxo.height = self.transactions[i].height
                    utxo.confirmed = self.transactions[i].confirmed
                    utxos.append(utxo)
        return utxos


    def advance_to_next_provider(self):
        if not self.provider_list:
            return
        
        newindex = (self.current_index + 1) % len(self.provider_list)
        self.provider_list[newindex].transactions = self.provider_list[self.current_index].transactions
        self.current_index = newindex
    
    def get_transaction_history(self):
        for address in self.addresses:
            txs = []
            ntransactions = -1  # Set to invalid value for the first iteration
            cycle = 1
            while ntransactions != len(self.transactions):
                if cycle > self.max_cycles:
                    raise NetworkException(f"None of the address providers are working after {self.max_cycles} tries")
                self.provider_list[self.current_index].transactions = txs
                self.provider_list[self.current_index].addresses = [address]
                try:
                    self.provider_list[self.current_index].get_transaction_history()
                    ntransactions = len(self.transactions)
                    txs = self.provider_list[self.current_index].transactions
                    break
                except NetworkException:
                    txs = self.provider_list[self.current_index].transactions
                    self.advance_to_next_provider()
                    cycle += 1
            self.transactions.extend(txs)
        return self.transactions